%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Implementation the GS (Global Store Module)  %%
% Author: Ilis Alouini, Mostafa Al-Metwally   %%
% @copyright UCL & Region Wallone              %%
% Read LICENSE                                 %%
% Last update:  15/01/01                       %%
% Documentation: http://www.ucl.ac.be/~ila/fts %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Class Client: Implements the deduction rules described in
% the paper:
% Reflective failure handling for collaborative applications 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Rules R1 -> R13: Non Fault Tolerant Model NFTM
%      R18,R5',R2',R7',R3': retrans operation NFTM
%      R19->R23: Recovery for lost messages
%      R24->R28: Election Rules
%      R29->R34: Building a new coherent store
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Representation of an object: (objectident,objectstate)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
functor
import
   System
   Fault
   Open
   OS
   Application
   ServerModule(serverClass:ServerClass) 
   StationaryObjectModule(newStatObj:NewStatObj)
  export
   ClientClass
define
   %%%%%%%%%%%%
   %% For Debug
   %%%%%%%%%%%%
   Debug=0
   proc {Show Message}
      if Debug==1
      then
	 {System.show Message}
      else
	 skip
      end
   end
   IncModelFlag=0
   %%%%%%%%%%%%%%%
   %% Client Class
   %%%%%%%%%%%%%%%
   class ClientClass  
      prop locking
      attr
	 server                 % Server Stationary Object
	 client                 % Client Stationary Object
	 user                   % User Object
	  
         %%%%%%%%%%%%%%%%%%%%%%%%	 
	 transobjects           % dictionary of selected objects (transid,list of objectident)	 
	 idtosavedobjects       % dictionary of (Transid,dictionary of object states (objectident,savedobjectstate)
	 objects                % dictionary of object states (objectident,coherentstate)	 
	 classofobjects         % dictionary of object classes (objectident,refobject#objectclass#initmethod)
	 updates                % dictionary of stored updates (transident,list of stored Objid#update)
	 pendingtrans           % dictionary of (transaction id,true/false)
	 objectlocks            % dictionary of (transaction id, list of locked objectident)
	 nametoint              % dictionary of (transaction id,integer)
	 locks                  % dictionary of (objid,client)
	 counters               % dictionary of (object id, counter of committed updates)
	 son                    % dictionary of (transaction id, son transaction id)
	 father                 % dictionary of (transaction id, father transaction id)
	 currenttransid:none    % Current transaction id
	 newlock:{NewLock $}    % to serialize election thread with current transaction thread
         lockrlstate:{NewLock $}% to serialize election thread with current transaction thread
	 endrecompute:true      % false if recomputation is in progress else true
	 connectstore:{NewLock $} % to synchronize connection protocol with transac. protocol
	 %committed              % dictionary of (transaction id,true/false) committe trans
	 recomputeset:nil       % Recomputed transaction set
	 tmplist:nil                % temp attributes
	 membership:nil         %list of othermembership (Username#Id|...)
	 
	 %%%%%%%%%%%%%%%%%%%%%%%%
	 %For debug
	 com:0 
         %%%%%%%%%%%%%%%%%%%%%%%%
	 localobjects:nil       % list of (refobject,objectident)
	 transset:nil           % active transactions
	 endedtransset:nil      % ended transactions
         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         % used for connection protocol
         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%      
	 lasttransid:none       %
	 connect:false          %
	 %%%%%%%%%%%%%%%%%%%%%%%%
	 %% election attributes
         %%%%%%%%%%%%%%%%%%%%%%%%
	 status:normal          % status of a client
	 leader                 % leader of election algorithm
	 newserver              % the new elected server
	 idelection:0           % identifier of the current election
	 aliveclients:nil       % list of alive clients while election
	 myident:0              % identifier of client
	 listmyident            % dictionary of (election id,list of clients ident)
	 identlist:nil          % list of other client ident  added by metwally
	 identclient:nil        % list of Clientref#clientid added by metwally
	 idsenders:nil          % added by metwally list of sender alstate 
	 procwatcherserver:proc{$ E C} skip end
	 watcherprocclient:nil  %lis of procwatcher#clientref
	
	 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	 %% new coherent store attributes
         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	 newlockedstore :nil        % List of (objectid#state#client) newstore after election 
	 newunlockedstore : nil      % List of (objectid#state) newstore after election 
	 ident:0                % unique identifier for rlstate message
	 identnewstore:0        % unique identifier for rlstate message
	 received:0             % counter
	 blocked:unit           % flag to block modification while election
	 lockedobjects          % list of locked objects  
	 unlockedobjects        % list of locked objects  
	 dictunlocked           % Dictionory of (objectid, List of states) 
	 listidstore:nil        % list of (Client Stationary Object, Ident Client) 
	 max                    % Used to compute recentstate
	 recentstate            % Used to compute recentstate	
	 failurekind:none       % Type  of server failure tempfail or permfail
	 
	 objwaitingref          % for waiting broadcast object copy
	 dictbroadcast
	 iaminupdate:false
	 %%%%%%%%%%%
	 % for debug
	 %%%%%%%%%%%
	 debug
	 i:0

	 leave:false   % for leave merge protocol
	 templist:nil  % to br reconsiderd again
	 mapedref:nil  % to be reconsidered again
	 oldserver
	 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Init the Client Class
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth init(Server Debug)
	 server <- Server 
	 debug <- Debug %Username
	 idtosavedobjects<-{NewDictionary $}
	 objects <-{NewDictionary $}
	 classofobjects<-{NewDictionary $}
	 transobjects <- {NewDictionary $}
	 objectlocks <- {NewDictionary $}
	 updates<- {NewDictionary $}
	 pendingtrans<- {NewDictionary $} 
	 nametoint <- {NewDictionary $}
	 counters <- {NewDictionary $}
	 listmyident <- {NewDictionary $}
	 dictunlocked <- {NewDictionary $}
	 locks <-  {NewDictionary $}  
	 son <- {NewDictionary $}
	 father <- {NewDictionary $}
	 %committed <- {NewDictionary $}
	 dictbroadcast <- {NewDictionary $}
	% identclient <- {NewDictionary $}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Rule R1 init transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth begin_trans(Transid)
	 {Wait @blocked} 
	 lock	    
	    {Show @debug#@myident#'client'#'begin_trans'}
	    {NewName ?Transid}
	    currenttransid <- Transid   
	    i <- @i+1
	    {Dictionary.put @nametoint Transid @i}	   
	    transset <- Transid|@transset
	    {Dictionary.put @pendingtrans Transid true}
	    {Dictionary.put @transobjects Transid  nil}
	    {Dictionary.put @objectlocks Transid  nil}
	    {Dictionary.put @idtosavedobjects Transid {NewDictionary $}}
	    {Show @debug#@myident#'client'#'begin_trans'#{Dictionary.get @nametoint Transid }}
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Rule R1 init recomputed transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth begin_transrecom(Transid) 
	 {Show @debug#@myident#'client'#'begin_transrecom'}
	 currenttransid <- Transid         
	 transset <- Transid|@transset	       	
      end      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Transform localref O to GlobalRef
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth objtoglobalref(O ?GlobalO)
	 fun {Togref Obj}
	    Objid in
	    try
	       {ForAll  @localobjects
		proc {$ O#Oid}
		   if O==Obj then Objid=Oid raise endforall end
		   else
		      skip
		   end
		end
	       }
	       {Show @myident#'client'#'internal_error(0)'}
	       {Application.exit 0}
	    catch endforall
	    then
	       skip
	    end
	    Objid % Obj exists in the store
	 end
      in
	 GlobalO={Togref O}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Rule R2 Apply M on Object
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth user_update(Transid Objectid Meth)
	 {Wait @blocked}
	 lock
	    lock @lockrlstate then	       
	       {self  user_update1(Transid Objectid Meth)}
	    end
	 end
      end
      meth user_update1(Transid Objectid Meth)
	Obj  State Objids Updates Dictstate Res ObjRes
      in 
	 %{self objtoglobalref(Obj Objectid)} % we should do something if this object is deleted!!  
	 {Dictionary.condGet @son Transid noentry Res}
	 if Res==noentry orelse {List.member Transid @recomputeset $}==true 
	 then
	   % Precondition: Obj is in the store this is very important condition     
	    if {List.member Transid  @transset}==true  %andthen {Dictionary.condGet @committed Transid noentry $}==noentry  
	    then	       
	       {Show @debug#@myident#'client'#'user_update'#{Dictionary.get @nametoint Transid $}#Meth}
	       %{self objtoglobalref(Obj Objectid)}
		  %{self getglobalref(Obj Objectid)}
	       {Dictionary.condGet @objects Objectid noentry ObjRes}
	       if ObjRes\=noentry then
		  %Site proberty
		  %{SiteProperty.get Objectid ?Obj}
		  {Dictionary.get @classofobjects Objectid Obj#_#_}
		  {Show @myident#objectfound#Obj}
		  try {Dictionary.get @transobjects Transid ?Objids}
		  catch _
		  then
		     {Show @myident#'GETERROR:dict:@transobjects'}
		     {Application.exit 0}
		  end
		  if {List.member Objectid Objids}==false 
		  then		  
		     {Dictionary.remove @transobjects Transid}
		     {Dictionary.put @transobjects Transid {List.append Objids  [Objectid] $}}
		  % request lock if not recomputed
		     {Show @myident#sendingrequestlock}
		     try {@server request_locks([Objectid] Transid @client#@myident)}
		     catch senderror % if send fails R18
		     then
			LL in
			{Show @debug#@myident#'client'#'request_locks'#'request_locksfails'}
			{self no_locks(Transid)}
			{Dictionary.keys @objects LL}
			{Show @myident#com#@com}
			{ForAll LL
			 proc {$ Oi}
			    O in
			    {Dictionary.get @classofobjects  Oi O#_#_}
			    %{Show 'lockfails:val:'#{O show($)}}   
			 end
			}
			raise endtrans end
		     end
		  else
		     skip
		  end
               % Save state
		  {Dictionary.get @idtosavedobjects Transid Dictstate}
		  if {Dictionary.condGet  Dictstate Objectid noentry $}==noentry then 
		     {Obj toChunk(?State)}
		     {Show @myident#'addedtoDictstate'}
	          % replace the entry (objectident,objectState) in the entry of idtosavedobjects 
		     {Dictionary.put Dictstate Objectid State}
		  else
		     skip
		  end
	       % Apply the method
		  {Show @myident#beforapplyMeth#Obj#Meth}
		  {Obj Meth}
		  {Show @myident#afterapplyMeth#Obj#Meth}
	       % Store the modification Meth
		  {Dictionary.condGet @updates Transid  noentry ?Updates}
		  if Updates==noentry then
		     {Dictionary.put @updates Transid [Objectid#Meth]} 
		  else
		     {Dictionary.remove @updates Transid}
		     {Dictionary.put @updates Transid {List.append Updates [Objectid#Meth] $}}
		  end
		  {Show @debug#@myident#'client'#'user_update_end'}
	       else
		  {Show @debug#@myident#'client'#'object deleted'#{Dictionary.get @nametoint Transid $}#Meth}
		  raise endtrans end
	       end 
	    else
	       %{Show {Dictionary.condGet @committed Transid noentry $}}
	       {Show @debug#@myident#'client'#'user_update_ignore'#{Dictionary.get @nametoint Transid $}#Meth}
	           % transaction  must terminates
	       raise endtrans end
	          % R12 ignore update Transid is killed  or already aborted
	       
	    end
	    
	 else
	    {Show @debug#@myident#recursive_update}
	    %{Application.exit 0}
	    {self user_update1(Res Objectid Meth)}
	    
	 end
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Rule R3 end transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth end_trans(Transid)
	 {Wait @blocked}
	 lock
	    lock @lockrlstate then	       
	       {self end_trans1(Transid)}
	    end
	 end
      end
      meth end_trans1(Transid)	
	 Objs Locks Res in

	 recomputeset <- {List.subtract  @recomputeset  Transid $} 
	 {Dictionary.condGet @son Transid noentry Res}
	 if Res==noentry
	 then
	    if {List.member Transid @transset $}==true
	    then
	       {Show @debug#@myident#'client'#'end_trans'#{Dictionary.get @nametoint Transid $}}	        
	       % Remove Trasnid from transset
	       transset <-{List.subtract @transset Transid $}	       
	       try {Dictionary.get @transobjects Transid ?Objs}
	       catch _ then {Show @myident#'Internal Error 2'#transobjects} {Application.exit 0}
	       end        
	       try {Dictionary.get @objectlocks Transid ?Locks}
	       catch _ then {Show @myident#'Internal Error 3'#objectlocks#Locks} {Application.exit 0}
	       end
	       if Objs==Locks 
	       then
		  if {IsFree @endrecompute $}==false then
		     {self all_locks(Transid)}
		  else
		     {self all_locks1(Transid)}
		  end
	       else	  
		  endedtransset <- Transid|@endedtransset
	       end	         
	    else
	       skip % R13 ignore en transaction aborted or already killed
	    end
	 else
	    {Show @debug#@myident#recursive_endtrans#Res#{Dictionary.get @nametoint Res $}}
	    %{Application.exit 0}
	    {self end_trans1(Res)}
	 end
      end      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Rule R4 abort transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth no_locks(Transid)
	 lock
	    {self no_locks1(Transid) }
	 end
      end
      meth no_locks1(Transid)  
	 lock @connectstore then
	    {Show @debug#@myident#'no_locks:'#{Dictionary.get @nametoint Transid $}}
	    %{Show abortsended#{Dictionary.get @nametoint Trid $}}	
	    %{@user abort(Transid)} 
	    %{Show @debug#cascading#{List.map {List.append @endedtransset @transset $} fun{$ Id} {Dictionary.get @nametoint Id $} end }}
	    {ForAll {List.append @transset @endedtransset $}
	     proc {$ Trid}
		Listobjid  Listobj Dictstate in		
                % Test if obj(Trid) inter obj(Transid) not empty
		try
		   
		   {ForAll {Dictionary.get @transobjects Transid $}
		    proc {$ Objid}
		       if
		          % The first  condition is necessary to avoid handling already aborted
		          % transactions
			  {Dictionary.condGet @pendingtrans Trid noentry $}==true			  
			  %%andthen % cascading abort
			  %% ({List.member Objid
			  %% {Dictionary.get @transobjects Trid $}  $}==true )                                                                         %%% RULE R4' : 
			  %orelse {Dictionary.condGet @father Trid noentry $}\= noentry )
		       then
			  {Show @debug#@myident#'no_locks_cascade:'#{Dictionary.get @nametoint Trid $}}
			  raise endno_locks end
		       else
			  skip
		       end
		    end
		   }
		catch endno_locks
		then
		   % roll back initial transaction
		   {Show @myident#'ROLLBACK'#{Dictionary.get @nametoint Trid $}}		      
		   try
		      {Dictionary.get @idtosavedobjects Trid ?Dictstate}
		       {Show @myident#'ROLLBACK:objects'#{Dictionary.get @transobjects  Trid $}#{Dictionary.entries Dictstate $}}
		      {ForAll  {Dictionary.get @transobjects  Trid $}
		       proc {$ Objid}
			  State Obj in
			  try 
			     {Dictionary.get Dictstate Objid ?State}			  
			     {Dictionary.get @classofobjects Objid Obj#_#_}
			     {Obj fromChunk(State)}
                             %%% execute action abort
			     %{self setupdate}
			     %{@user executeactionabort(Objid)}
			     %{self unsetupdate}		     
			     {Dictionary.remove @objects Objid}  % To VERIFY 
			     {Dictionary.put @objects Objid State}% To VERIFY
			  catch _ then
			     skip
			  end
		       end
		      }
		   catch X % already rollbacked with updates
		   then {Show @myident#X} skip
		   end
		 
		   {Dictionary.get @transobjects Trid ?Listobjid}	                
		   {Dictionary.get @objectlocks Trid ?Listobj}
		   {ForAll Listobj
		    proc {$ Objid}
		       {Dictionary.remove @locks  Objid}
		    end
		   }	    
		   try {@server release_locks(Listobj Trid @myident)}
		   catch  senderror %R...
		   then
		      skip
		   end
		   {Dictionary.remove @objectlocks Trid}
                   % Remove  Transid from transset
		   transset <- {List.subtract @transset Trid $}	
                   % Remove Transid from  endedtransset
		   endedtransset <- {List.subtract @endedtransset Trid $}	       
		   {Dictionary.remove @pendingtrans Trid} 
		   {Dictionary.put @pendingtrans Trid false}
	           % remove stored updates
		   {Dictionary.remove @updates Trid}
		   {Show @myident#'REcursivecall to no_locks'}
		   if {IsFree @endrecompute $}==false then   
		      {self no_locks(Trid)}
		   else
		      {self no_locks1(Trid)}
		   end
		   % Notify user
		   {@user abort(Transid)}
		end		    
	     end
	    }	    
	end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Rule R5 commit transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth all_locks(Transid)
	 lock
	    {self  all_locks1(Transid)}
	 end
      end
      meth all_locks1(Transid)
	 fun {GetStates ListObjects}
	    State Obj in
	    case ListObjects of
	       nil then nil
	    [] Objid|Tail then
	       {Dictionary.get @classofobjects Objid Obj#_#_}
	       {Obj toChunk(?State)} % FAUX 
	       Objid#State|{GetStates Tail}
	    end
	 end
      in
	 {Wait @blocked}
	 lock @connectstore then
	    lock @lockrlstate then
	       Trid  Listobj Dictstate Dic Liststates in 
	       {Show @debug#@myident#'client:all_locks:'#{Dictionary.get @nametoint Transid $}}
	       com <- @com+1
	       %{Dictionary.put @committed Transid true}
	       {Dictionary.remove @pendingtrans Transid} 
	       {Dictionary.put @pendingtrans Transid false}
	       % notify user
	       %{@user commit(Transid)}
	       
	       if IncModelFlag==1 then
		  try {@server update({Dictionary.get @updates Transid $} {Dictionary.get @nametoint Transid $} @client)} 
		  catch senderror then skip% if send fails R..
		  [] E then {Show @myident#'ERROR'#E} {Application.exit 0} skip % noupdates
		  end
	       else
		  skip
	       end
	       %%%%%%%%%%%%%%%  AJOUT
	       {Dictionary.get @idtosavedobjects Transid ?Dictstate}
	       {Dictionary.entries Dictstate ?Liststates}
	       {Show @myident#'Commit:liststates:'#Liststates}
	       Dic={NewDictionary}
	       {ForAll Liststates
		proc {$ X}
		   Objid State  Obj Class Init in
		   if X\=nil then
		      X=Objid#State
		      {Show insideforall}
		      {Dictionary.get @classofobjects Objid _#Class#Init}
		      {Show after}
		      Obj={New Class Init}
		      {Obj fromChunk(?State)}
		      {Dictionary.put Dic  Objid Obj#State}
		   else
		      skip
		   end
		end
	       }
	       % apply updates
	       {Show @myident#'applyupdates:'#{Dictionary.get @updates Transid $}}
	       {ForAll {Dictionary.get @updates Transid $}
		proc {$ M}
		   Objid O S U in
		   if M\=nil then
		      M=Objid#U
		      {Dictionary.get Dic  Objid  O#S}
		      {O U}
		      {Dictionary.remove Dic  Objid}
		      {Dictionary.put Dic Objid  O#{O toChunk($)}}
		   else
		      skip
		   end
		end
		}		      
%%%%%%%%%%%%%%%% FIN AJOUT
	       {Show @myident#'transobject:'#{Dictionary.get @transobjects  Transid $}}
	       {ForAll {Dictionary.get @transobjects  Transid $}
		proc {$ Objectid}
		   State in
		   {self increment(Objectid)}
                   % replace the entry (objectident,objectState) in the dictionary objects
		   {Dictionary.get Dic Objectid _#State}
		   {Dictionary.remove @objects Objectid}
		   {Dictionary.put @objects Objectid State}
		end
	       }
	       if IncModelFlag==0 then  % it could send to all members from her not from server side.
		  try {@server update({Dictionary.entries Dic $} {Dictionary.get @nametoint Transid $} @client)} 
		  catch senderror then skip% if send fails R..
		  [] E then {Show @myident#'ERROR'#E} {Application.exit 0} skip % noupdates
		  end
	       else
		  skip
	       end
	       %notify user
	       {@user commit(Transid)}
	       {Dictionary.remove @transobjects Transid}	    
	       lasttransid <- Transid
	       {Dictionary.remove @updates Transid}
	       % Rule R5'
	       {Dictionary.condGet  @son Transid noentry ?Trid}
	       if Trid==noentry then
		  {Dictionary.get @objectlocks Transid ?Listobj}
		  {ForAll Listobj
		   proc {$ Objid}
		   {Dictionary.remove @locks  Objid}
		   end
		  }
		  try {@server release_locks(Listobj Transid @myident)}
		  catch  senderror
		  then skip % R19	  
		  end
		  {Dictionary.remove @objectlocks Transid}
	       else
		  skip % son transaction will release locks
	       end
	      
	       {Dictionary.removeAll Dic} 
	       {Dictionary.remove @idtosavedobjects Transid}
	       transset <- {List.subtract @transset Transid $}
	       {Dictionary.remove @pendingtrans Transid} 
	       {Dictionary.remove @nametoint Transid}
	    end
	 end	 
      end

      meth getstate(N ?State1)
	 %{Wait @blocked}
	 %lock @connectstore then
	 %   lock @lockrlstate then
	 lock
	    {Dictionary.get @objects N State1}
	    {Show @myident#getstate#State1}
	 %   end
	 end
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Rule R6 update state: call from server
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth update(Methods)	 
	 lock    
           % Get {List.append  @endedtransset @transset $} without retrans id's
	    fun {NewSet L}
	       case L of
		  nil then nil
	       []
		  X|Tail then
		  try 
		     
		     {Dictionary.get @father X _}				     
		     {NewSet Tail}
		  catch _  % X has no father
		  then	     
		     X|{NewSet Tail}
		  end
	       end
	    end
	     LL
	    
	 in
	    iaminupdate <- true
	    {Show @debug#@myident#'client:'#'updates:'#Methods}
	    recomputeset <-  {List.append {List.reverse @endedtransset $}{List.reverse @transset $} $}
	    {Show @myident#afterrecompute#recomputeset}
	   tmplist <-{NewSet {List.append {List.reverse  @endedtransset $}{List.reverse @transset $} $}}
	    {Show @myident#afternewl#@tmplist}
	    {ForAll Methods
	     proc {$ X}
		Objid in
		if X\=nil then
		   X=Objid#_	   
		   {ForAll {List.append @transset @endedtransset $}
		    proc {$ Trid}
		       Dictstate in
		       %if true==true  %{List.member Objid {Dictionary.get @transobjects Trid $}}==true  %true==true 
		       if {List.member Objid {Dictionary.get @transobjects Trid $}}==true  %true==true 
		       then
		          % roll back states
			  {Show @debug#@myident#'client:'#'updates:rollbacktransaction'#{Dictionary.get @nametoint Trid $}#{Dictionary.get @transobjects Trid $}}
			  try 
			     {Dictionary.get @idtosavedobjects Trid ?Dictstate}	    
			     {ForAll {Dictionary.get @transobjects Trid $}
			      proc {$ Oid}
				 State Obj  in
				 try 
				    {Dictionary.get Dictstate Oid ?State}
				    {Dictionary.get @classofobjects Oid Obj#_#_}
				    {Obj fromChunk(State)}
				    %{Show 'API:'#{Obj show($)}}
				 catch _ % Object is not in the set of transobjects (recomputed)
				 then skip
				 end				 
			      end
			     }
			  catch E % already removed
			  then {Show @myident#E}
			  end
			  % remove stored updates and trid
			  {Show @myident#beforremoveinsideforall}
			  {Dictionary.remove @updates Trid}
			  % remove saved states
			  {Dictionary.remove @idtosavedobjects Trid}	  
			  {Dictionary.put @idtosavedobjects Trid {NewDictionary $}}
			  endedtransset <- {List.subtract @endedtransset Trid $}
			  transset <- {List.subtract @transset Trid $}    
		       else
			  skip
		       end
		    end
		    }
		else
		   skip	   
		end
	     end
	    }
	    {Show @myident#afterfristforall}
	    % Apply received methods or put new states
	    {ForAll Methods
	     proc {$ X}
		M Objid Obj State in
		try
		   if X\=nil then
		      if IncModelFlag==1 then
			 X=Objid#M
		      else
			 X=Objid#(_#M)
			 {Show @myident#M}
		      end
		      try
			 {Dictionary.get @classofobjects Objid Obj#_#_}
		      catch E then {Show @myident#'objecterror'#E} end
		      {Show @myident#'objectrefrencefound'#Obj}
		      %{Show 'method'#M}
		      if IncModelFlag==1 then
			 %{@user apply(Objid M)}
			 {Obj M} % this hould be applied at user not her!!!
			 % if it is applied her we will have a reserouce problem
			 {Show @myident#'method'#M}
		      else  % M is a state
			 {Show @myident#beforchunk}
			 try
			    {Obj fromChunk(M)}
			 catch E then
			    {Show @myident#'error'#E}
			 end
		      end
		      %notify user 
		      %{@user executeactionupdate(Obj)}
		      {@user executeactionupdate(Objid)}
		      %if {List.member Objid @tmp}==false then
		      {self increment(Objid)}
		      %else
			% skip
		      %end
		      %tmp <- Objid|@tmp
                      % replace the entry (objectident,objectState)
		      % in the dictionary objects
		      {Show @myident#afterincrement}
		      {Dictionary.remove @objects Objid}
		      {Obj toChunk(?State)}
		      {Dictionary.put @objects Objid State}			      
		   else
		      skip
		   end
		catch X then
		   {Show @myident#X#M}
		   {Application.exit 0}
		end
	     end
	    }
	    {Dictionary.keys @objects LL}
	    {Show @myident#com#@com#LL}
	    % {ForAll LL
% 	     proc {$ Oi}
% 		O in
% 		{Dictionary.get @classofobjects  Oi O#_#_}
%   		  % {Show 'val in update'#{O show($)}}		   
% 	     end
% 	    }
	   % tmp <- nil
	    % ignore init user_update message from now until receive recomputedone
	    %{Show @debug#@myident#'client:'#'recompute'#{List.map NewL fun{$ Id} {Dictionary.get @nametoint Id $} end }}
	    
	    if @tmplist \= nil  then
	       endrecompute <- _    
	       {@user recompute(@tmplist)}
	       {Wait @endrecompute}
	       recomputeset <- nil
	       {Show @debug#@myident#'client:'#'recomputeFIN'}
	       
	    else skip end
%%% execute action
	    iaminupdate <- false
	    {Show @myident#'OUTOFLOCK1'}
	    {Show @myident#'OUTOFLOCK1'}
	    recomputeset <- nil
	    tmplist <-nil
	 end  % end of lock
	 {Show @myident#'OUTOFLOCK2'#@tmplist}
	 %recomputeset <- nil
	 
      end % end of method body
      meth endrecom
	 {Show @myident#'client:ENDRECOMPUTE'}
	 @endrecompute=true
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Rule R7 locks from server 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth serv_all_locks(Transid Sel)
	 lock
	   {self serv_all_locks1(Transid Sel)} 
	 end
      end    
      meth serv_all_locks1(Transid Sel)
	  
          
	    Res Trid in
	    {Dictionary.get @pendingtrans Transid Res}
	   
	    if Res==true then	   
	       Ls Lsobj in
	       
	       {Dictionary.get  @objectlocks Transid ?Ls}
	       {Dictionary.remove @objectlocks Transid}
	       {Dictionary.put @objectlocks Transid {List.append Ls Sel $}}
	       % update locks dictionary
	       {ForAll Sel
		proc {$ Objid}
		   {Dictionary.put @locks  Objid @client}
		end
	       }
	       {Dictionary.get  @transobjects Transid ?Lsobj}
	       {Show @debug#@myident#'client'#'serv_all_locks'#{Dictionary.get @nametoint Transid $}}	 
	       if {List.member Transid @endedtransset $}==true andthen
		  Lsobj=={Dictionary.get @objectlocks Transid $}
	       then
		  {self all_locks(Transid)}	  
		  endedtransset <- {List.subtract @endedtransset Transid $}	  
	       else
		  skip  
	       end
	       % R7'
	      
	       {Dictionary.condGet  @son Transid noentry ?Trid}
	       if Trid\=noentry
	       then
		  {Show @debug#@myident#'client'#'serv_all_locks:recursive'}
		  {self serv_all_locks(Trid Sel)}
	       else
		  skip
	       end				       
	    else
	    % Rule R11 ignore received locks	       
	       try {@server release_locks(Sel Transid @myident)}
	       catch  senderror
	       then skip % R19	  
	       end	       
	    end
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Rule R8 locks from server 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth serv_no_locks(Transid Sel)
	 lock
	   {self serv_no_locks1(Transid Sel)} 
	 end
      end
      meth serv_no_locks1(Transid Sel)
	 Res  LL in
	 {Dictionary.get @pendingtrans Transid Res}    
	 if Res==true then
	    {Show @myident#'serv_no_locks:begin_no_locks'} 
	    if {IsFree @endrecompute $}==false then         
	       {self no_locks(Transid)}
	    else
	       {self no_locks1(Transid)}
	    end
	      {Show @myident#'serv_no_locks:end_no_locks'} 
	       if {List.append @transset @endedtransset $}\= nil
	       then
              {Show @myident#set_afterno_locks#{List.map {List.append    @endedtransset @transset $} fun{$ Id} {Dictionary.get @nametoint Id $} end }}
		 %{Application.exit 0}
	       else
		  skip
	       end
	       {Dictionary.keys @objects LL}
	       {Show @myident#com#@com}
	       % {ForAll LL
% 		proc {$ Oi}
% 		   O in
% 		   {Dictionary.get @classofobjects  Oi O#_#_}
%   		   {Show 'val:'#{O show($)}}   
% 		end
% 	       }
               
               % remove the Transid from transobjects
	       %{Dictionary.remove @transobjects Transid}
	      
	    else
	       skip %Rule R10
	    end
	
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Rule R9 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth waitlocks(Transid Sel)
	 {Wait @blocked}
	 lock
	    {self waitlocks1(Transid Sel)}
	 end
      end
      meth waitlocks1(Transid Sel)	 
	 Selid Result LastSelid
	 Objs
	 fun {Sub L LS}
	    case L of
	       nil then nil
	      []  X|Tail then {List.append {List.subtract LS  X $} {Sub Tail LS} $}
	    
	    end
	 end
      in
	
	 if {List.member Transid @transset $}==true 
	    then
	       {List.map Sel
		fun {$ Obj}
		   Lobj in
		   try 
		      {Obj getglobalref(Lobj)}
		      Lobj
		   catch _ then
		      raise noobjectinstore % raise exception no legal object
		      end
		   end
		end
		?Selid
	       }
               %%%%% if object in objectslocks then remove
	       %%%% LastSelid: objects that have not yet the locks
	       
               LastSelid={Sub Selid {Dictionary.get  @objectlocks Transid $} }
	       {Dictionary.get @transobjects Transid ?Objs}
	       {Dictionary.remove @transobjects Transid}
	       {Dictionary.put  @transobjects Transid {List.append Objs LastSelid $}}
	       
	       if LastSelid\= nil then
		  {@server waitlocks(Transid LastSelid  Result @client)}
		  {Show @myident#'WAITLOCKS'#LastSelid}
		  {Wait Result}       
		  {Show @myident#'WAITLOCKS'#Result#LastSelid}
	          % Synch=true Ok locks obtained else no abort transid
		  if Result==true then
		     if {IsFree @endrecompute $}==false then  
			{self serv_all_locks(Transid Selid)}
		     else
			{self serv_all_locks1(Transid Selid)}
		     end
		  else
		     if {IsFree @endrecompute $}==false then  
			{self serv_no_locks(Transid Selid)}
		     else
			
			{self serv_no_locks1(Transid Selid)}
		     end
		  end
	       else
		  skip
	       end
	    else
	       skip
	    end
	
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Rule R18 retrans operation
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth retrans(?Rtrid)
	 if {IsFree @endrecompute $}==false
	 then
	    lock
	       {self retrans1(?Rtrid)}
	    end
	 else
	    {self retrans1(?Rtrid)}   
	 end
      end
      meth retrans1(?Rtrid)	 
	 NewTransid Objs Locks 
	 
	 Trid Transid in
	 
	 {self getcurrenttransid(Transid)}	 
	 recomputeset <- {List.subtract  @recomputeset  Transid $} 
	 {Show @myident#'client:retrans:recomputeset'#{Dictionary.get @nametoint Transid $}#@recomputeset}
	 if {IsFree Rtrid $}==true then  
	    if {List.member Transid @transset $}==true then
	       {Show @myident#'client:retrans'#{Dictionary.get @nametoint Transid $}#@transset }
	       try
		  {Dictionary.get @transobjects Transid ?Objs}
		  {Dictionary.get @objectlocks Transid ?Locks}
	       catch _
	       then
		  {Show xxxxxx} {Application.exit 0}
	       end
	       {Show @myident#'client:retrans'#{Dictionary.get @nametoint Transid $}}
	       if Objs==Locks 
	       then
		  {self all_locks(Transid)}
	       else
		  endedtransset <- Transid|@endedtransset 
	       end
	       transset <-{List.subtract @transset Transid $}       
	      		 
	       {Dictionary.condGet  @son Transid noentry ?Trid}
	       if Trid==noentry then
		  {NewName ?NewTransid}
		  Rtrid=NewTransid
		  transset <- NewTransid|@transset
		  i <- @i+1
		  {Dictionary.put @nametoint NewTransid @i}	   	  
		  {Dictionary.put @pendingtrans NewTransid true}
		  {Dictionary.put @son Transid NewTransid}
		  {Dictionary.put @father  NewTransid Transid}
		  {Dictionary.put @transobjects  NewTransid Objs}
		  {Dictionary.put @objectlocks  NewTransid Locks}
		  {Dictionary.put @idtosavedobjects NewTransid {NewDictionary $}} % faux
		  {Show @myident#'client:retrans:created'#Transid}
		  
	       else		  
		  transset <- Trid|@transset
		  Rtrid=Trid	
	       end
	       
	    else

	       
		  {Show @myident#'client:retrans ignored'}
                % ignored retrans -> aborted 
	    
		  Rtrid={NewName}
		  i <- @i+1
		  {Dictionary.put @nametoint Rtrid @i}	   	  
		  {Dictionary.put @pendingtrans Rtrid  false}
		  {Dictionary.put @son Transid Rtrid}
		  {Dictionary.put @father  Rtrid Transid}
		  {Dictionary.put @transobjects  Rtrid nil}
		  {Dictionary.put @objectlocks  Rtrid nil}
		  {Dictionary.put @idtosavedobjects Rtrid {NewDictionary $}}
	       
		  thread {@user abort(Rtrid)} end
		  {Show @myident#'client:retrans ignored'}

	    
	    
	    end

	 else
	    if {IsFree @endrecompute $}==true then
	       endedtransset <- Transid|@endedtransset 
	       transset <-{List.subtract @transset Transid $}     
	       transset <- Rtrid|@transset % recomputation
	    else
	       skip 
	    end
	    
	 end
	
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% count updates on each object
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
      meth increment(Oid)
	 Counter in
	 {Dictionary.get @counters Oid  ?Counter}   
	 {Dictionary.remove @counters Oid }
	 {Dictionary.put @counters Oid Counter+1}
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Get current Transid
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
      meth  getcurrenttransid(?Tid)
	 Tid=@currenttransid
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%  Rules for Fault Tolerance 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Rule R24 action after server failure 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth watcheronserver(Failurekind)
	 lock
	    Lsobj Ls in
	    if @connect==true then
	       failurekind <- Failurekind   
	       {Show @debug#@myident#'client'#':watcheronserver'#@server}
	       status <- selection
	    % wait end of recompute
	       {Wait @endrecompute}
	       {ForAll  @endedtransset  
		proc{$ Transid}		
		{self no_locks(Transid)} % abort transaction 
		end
	       }		   
	       {ForAll @transset  % one current transaction 
		proc{$ Transid}
		   {Dictionary.get  @transobjects Transid ?Lsobj}
		   {Dictionary.get  @objectlocks Transid ?Ls}
		   if Lsobj\=Ls then
		      {self no_locks(Transid)}
		   else		   
		   {Show @debug#@myident#'client'#'watcheronserver:'#Transid#'notaborted'}
		      skip
		   end
		end
	       }
	       {self initiatselection()}
	       {Show @debug#@myident#'client'#'watcheronserver'#@server}
	    else
	       {Show @myident#'i am already disconnected (inside failure'#@myident}
	    end
	 end
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule R24 init an election
      %%%%%%%%%%%%%%%%%%%%%%%%%%% 
      meth initiatselection()
	 lock
	    {Show @debug#@myident#'initiatselection:'#@myident}
	    {Show @debug#@myident#'clientselection:aliveclients'#@aliveclients#@identlist}
	    %status <- selection
	    %leader <- @myident	   
	   % newlockedstore<- nil
	    %received <- 0
	    % block clients until the end of selection + new state of gs see paper
	    blocked <- _
	    if @aliveclients==nil
	    then
	       % one alive client : no election
	       %ident <- @ident+1
	       %{self rlstate(@ident @client @failurekind)}
	       %status <- preorg
	       
	       leader <- @myident
	       status <- preorg
	       {Show @debug#@myident#'client'#@myident#':I am leader'}
	       {self rlstate(@ident @client @failurekind)}
	       {Show @debug#@myident#'SELFINIT_RLSTATE'}
	    else
	       {self selectleader()}
 	    end
	    {Show @debug#@myident#'client:initiatselection:end'}
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule 25 receiving myident from alive clients
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule 25bis
      % Test if all messages of the current election arrived
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
      meth selectleader()
      % Listmax: compute maximum of a list Li
	  fun {Listmax Li}
	    Y in
	    case Li of
	       nil then 0
	    [] X|L then
	       Y={Listmax L}	       
	       if X < Y    % it was <
	       then
		  Y
	       else		  
		  X
	       end
	    else
	       0  
	    end
	  end
	  proc{Getidref Id Li ?Ref}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  Ref=H.1
	       else
		  {Getidref Id Tail ?Ref}
	       end
	    end
	 end
	  
	 Meni Leaderref
      in
	 {Show @myident#'selectleader'#{List.length @identlist  $}#{List.length @aliveclients $}}
	 lock
	    if @identlist== nil then skip
	    else
	       if {List.length @identlist  $}=={List.length @aliveclients $} then
	       {Show @myident#'listlengths'#{List.length @identlist  $}#{List.length @aliveclients $}}
	       %if true==true then
		  {Show @debug#@myident#'client this is my ident:='#@myident}
		  Meni={List.nth @identlist {List.length @identlist} $}
		  if Meni > @myident then
		     leader <- @myident
		     status <- preorg
		     {Show @debug#@myident#'client'#@myident#':I am leader'}
		     %{self dummy()}
		     %the leader must sen to all alive clients a message to know the failed client
 		     {self rlstate(@ident @client @failurekind)}
		     {Show @debug#@myident#'SELFINIT_RLSTATE'}
		  else
		     leader <- Meni
		     {Show 'leader'#@leader#'myident'#@myident#@identlist}
		     Leaderref={Getidref Meni @identclient $}
		     {self rlstate(@ident Leaderref @failurekind)}
		     skip 
		  end
	       else	  
		  {Show @myident#notequal}
		  skip		  
	       end
	    end
	 end
      end

      meth dummy()
	 {ForAll @aliveclients
	  proc {$ C}
	     if C\=@client then
		try
		   {Show 'atleader '#@myident}
		   {C dummymsg()}
		catch senderror
		then
		   {Show @myident#'this client is dead'}
		end
	     end
	  end
	 }
      end

      meth dummymsg()
	{Show 'this my id'#@myident}
      end
		
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Rule R26 watcher at C on alive clients
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth watcherproc(C)
	 Clintid MemUser
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	 proc{Memberuser Id Li ?MemUser}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  MemUser=H.1
		  %{Show 'usermem='#MemUser}
	       else
		  {Memberuser Id Tail ?MemUser}
	       end
	    end
	 end
      in
	 if @connect==true then 
	    if {List.member C @aliveclients $}
	    then
	       Clintid={Getident C @identclient $}
	       {Show @myident#'crashed client'#Clintid}
	       MemUser={Memberuser Clintid @membership $}
	       if {String.is MemUser $}==true
	       then
		  skip
	       else
		  skip
	       end
	       {Show'this is user'#@myident#MemUser#@membership#Clintid}
	       aliveclients <- {List.subtract @aliveclients C $}
	       identlist <- {List.subtract @identlist Clintid $}
	       identclient <- {List.subtract @identclient C#Clintid $}
	       membership <- {List.subtract @membership MemUser#Clintid $}
	       {Show'failed clientid: '#Clintid#'myidentis'#@myident#'my identlist'#@identlist#@leader#@membership}
	       {Show @myident#'listlengths'#{List.length @identlist  $}#{List.length @aliveclients $}}
	       % try    
% 		  {@server ping()}
% 	       catch senderror
% 	       then
% 		  {Show 'server reference is dead'}
% 	       end
	       if @status \= normal then
		  if @leader==@myident then
		     if {List.member Clintid @idsenders $} then
			{Show @myident#'i am the leader apply recievedfromclient'#@leader#Clintid#@myident}
		     %{self recievedfromclient(C Clintid)}
		     else
			{Show @myident#'somthing'}
			{self recievedfromclient(C Clintid)}
		     end
		  else
		     {Show @myident#'i am not the leader'}
		  end
		  if @leader == Clintid then
		     {Show @myident#'i will select another leader'#@leader#Clintid#@myident}
		     {self initiatselection()}
		  end
	       else
		  {@user executeactionmembership(@membership)}
		  %{@user modifymembership(@membership)}
		  {Show @debug#@myident#@membership#'insidewatcher'#@status}
	       end
	    else
	       {Show @myident#'this is already not a member'}
	    end
	 else
	    {Show @myident#'i am already not connected (inside watcherproc)'#@myident}
	 end
      end
     
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Methods  for building a new coherent store
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule 29 init request locked objects
      % executed at leader
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule 30 request locked and unlocked objects
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth rlstate(Ident C Failtype)
	 LL %Lid
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	  in
	 lock @lockrlstate then            
            % to avoid deadlock in update and end_trans variable blocked ? to comment !?
	    Liststates in	   
	    if Failtype==nofail then
                 % merge in progress
	       thread
		  skip
		  %{@user block()}
	       end
	    else
	       skip
	    end
	    lockedobjects <- nil
	    unlockedobjects <- nil
	    {Show @debug#@myident#'client:receivedrlstate'#Failtype#rlstate(Ident C Failtype)}
	    % get all locked states a modifier    
	    {ForAll {Dictionary.keys @locks $}  
	     proc {$ Objid}
		lockedobjects <- Objid#{Dictionary.get @objects Objid $}#@client|@lockedobjects          
	     end
	    }
            % Get unlocked objects of the store	    
	    {Dictionary.entries @objects ?Liststates}
	    {Show @myident#'liststates:'#Liststates}
	    {ForAll Liststates
	     proc {$ X}
		Objid State  in
		if X\=nil then
		   X=Objid#State
		   try {Dictionary.get @locks Objid _}
		   catch _ then
		      unlockedobjects <- Objid#State#{Dictionary.get @counters Objid $}|@unlockedobjects
		      
		   end		   
		else
		   skip
		end
	     end
	    }
	    % Send answer to leader C
	    %identnewstore <-  @identnewstore + 1
	    try
	       {C  alstate(@lockedobjects @unlockedobjects Ident @myident @client @identnewstore)}
	       {Show @debug#@myident#'client:receivedrlstate:sentok'#alstate(@lockedobjects @unlockedobjects Ident @myident @client @identnewstore)}
	       %lockedobjects<- nil % for solving memory leak
	       %unlockedobjects<- nil % for solving memory leak
	    catch senderror
	    then
	       {Show 'newleader is dead'#@myident} % select another one!
	       %{self watcherproc(C)} % add to solve watcher bug!!!!!!if good watcher no need fo it!!!!
	    end
	 end
	 {Show @debug#@myident#'client:receivedrlstate:OUT'}	
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule 31, 32 reveive locked & unlocked objects 
      % This method is only executed at client leader
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%      

      meth alstate(Lockedstore Unlockedstore Nalstate SenderId Client Nnewstore)
	 % Nalstate: identifier of message rlstate
	 % Nnewstore: identifier of message newstore
	 % Client : client that sends the newstore 
       
	 {Show   @debug#@myident#'client:receivedstore'#Lockedstore#Unlockedstore#@ident#Nalstate#SenderId}
	 if @ident==Nalstate % valid message
	 then
	    % in case this client i did not know him because of server failuer
	    if {List.member SenderId @idsenders $} == false
	    then
	       listidstore <- Client#Nnewstore |@listidstore 	      
	       newlockedstore <- {List.append Lockedstore @newlockedstore $} 
	       received <- @received+1
	       idsenders <- SenderId|@idsenders
	       % Split the received unlocked store to dictionary
	       newunlockedstore <- nil
	       {ForAll Unlockedstore
		proc {$ X}
		   Liststates Objid State Counter in
		   if X \= nil
		   then
		      X=Objid#State#Counter
		      {Dictionary.condGet @dictunlocked Objid noentry ?Liststates}
		      if Liststates== noentry then 
			 {Dictionary.put @dictunlocked Objid  State#Counter|nil}
		      else
			 {Dictionary.remove @dictunlocked Objid}
			 {Dictionary.put @dictunlocked Objid State#Counter|Liststates}
		      end
		   else
		      skip
		   end
		end
	       }
	       % All requests for all clients are received	
	       {Show @debug#@myident#'client:receivedcounter'#@received#@identlist} 

	       {self recievedfromclient(Client SenderId)}

	    else
	       {Show @myident#'this sender gives me its state before'}
	    end
	 else
	    {Show @myident#'notvalid rlstatemessage'} % R34
	 end
      end
      meth recievedfromclient(Client SenderId)
	 %%%% added to solve watcher bug!!!!!!%%%%%
	 Clintid MemUser Leaderref
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	 proc{Memberuser Id Li ?MemUser}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  MemUser=H.1
		  %{Show 'usermem='#MemUser}
	       else
		  {Memberuser Id Tail ?MemUser}
	       end
	    end
	 end
	 proc{Getidref Id Li ?Ref}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  Ref=H.1
	       else
		  {Getidref Id Tail ?Ref}
	       end
	    end
	 end
	 %%%%% end %%%%%
	 proc {Findistore Idc Li ?Y}
	    try
	       {ForAll Li
		proc {$ I}
		   if I \= nil
		   then		   
		      case I of !Idc#X then  Y=X raise endforall end
		      else
			 skip
		      end
		   end
		end
	       }
	    catch endforall then
	       skip
	    end
	 end
	 
	 ListStates  Leaderref
	 %Pid F %FProc
      in
	 if @received >= {List.length @aliveclients $}+1 %modification it was +1
	 then	  
	    {Show @debug#@myident#'at client:Global Locked objects'#@newlockedstore }
	    
                  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                  % Get recent state for all unlocked objects
                  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    {Dictionary.entries @dictunlocked ?ListStates}
	    {Dictionary.removeAll @dictunlocked}
	    newunlockedstore <- nil
	    {ForAll ListStates
	     proc {$ X}		     
		Objid List in
		if X \= nil
		then
		   X=Objid#List
		   
		   max  <- ~1
		   {ForAll List
		    proc {$ Y}
		       State Counter in
		       if Y \= nil then
			  Y=State#Counter
			  
			  if Counter > @max
			  then
			     max <- Counter
			     
			     recentstate <- State
			  else
			     skip
			  end
		       else
			  skip
		       end
		    end
		   }
		   newunlockedstore <- Objid#@recentstate|@newunlockedstore
		else
		   skip
		end
	     end
	    }
	    {Show @failurekind#@myident}
	    if @failurekind\=nofail then	  	  		    
	       @newserver=_
	       {Show @myident#'creation'#@newserver}
	       {NewStatObj ServerClass init(@newserver) @newserver _ _}
            %%% Store emulator pid 
	       % try
% 		  Pid={OS.getPID}
% 		  F={New Open.file init(name:'/tmp/pidserver' flags:[write create])}      
% 		  {F write(vs:Pid)}
% 		  {F close}
% 	       catch
% 		  E
% 	       then
% 		  {Show 'ErrorCreateFile'#E}
% 	       end
	       if @failurekind==tempfail then
                        % Only the leader will execute initiate_merge
		  {@newserver initiate_merge(@server)}
	       else
		  skip
	       end		      
                     % Build clientset at server
	       aliveclients <- {List.append @aliveclients [@client] $}
	       identlist <- {List.append @identlist [@myident] $}
	       identclient <- {List.append @identclient [@client#@myident] $}
	       membership <- {List.append @membership [@debug#@myident] $}
	       {Show @myident#'before setclient'#@identlist#@identclient}
	       {@newserver setclientset(@aliveclients @identlist @identclient @membership @client @newserver)}
		  % remove yourself
	       identlist <-{List.subtract @identlist @myident  $}
	       aliveclients <-{List.subtract @aliveclients @client  $}
	       identclient <- {List.subtract @identclient @client#@myident  $}
	       membership <- {List.subtract @membership @debug#@myident $}
                     % Build lock table attribute at server
	       {@newserver buildlocks(@newlockedstore)}	       	         
	             % Broadcast newstore to other clients
	       {ForAll @aliveclients
		proc {$ C}
		   {Show @debug#@myident#'atclient:sendnewstoreid'#{Findistore C @listidstore $}}
		   try
		      {C newstore(@newunlockedstore @newlockedstore @newserver {Findistore C @listidstore $})}
		   catch senderror
		   then
		      skip
		   end
		end
	       }		  
	             % Set locally the newstore
	       {Show @debug#@myident#'atclient:sendnewstoreid'#@identnewstore }  
	       {self newstore(@newunlockedstore @newlockedstore @newserver @identnewstore)}
	       %listidstore <- nil
	       %newlockedstore <- nil
	       %newunlockedstore <- nil
	      % FProc='server is now in your site'
	      %{@user executeactionrcovery}
	    else
		     % nofail:send the store to the server
	       {@server store(@newunlockedstore @newlockedstore)}
	    end
	 else
	    {Show @myident#'wait for next rlstate message'#{List.length @identlist  $}#{List.length @aliveclients $} }
	 end
%%%%%%%%%%%%%% added to solve watcher bug!!!!!%%%%%%%%%%%%
	 if @leader==@myident 
	 then
	    skip
	 else
	    if @leader==nil then skip
	    else
	       {Show @myident#'I should be the leader'#{List.length @identlist  $}#{List.length @aliveclients $}}
	       %Leaderref={Getidref @leader @identclient $}
	       %{self watcherproc(Leaderref)}
	      %  if {List.member Leaderref @aliveclients $}
% 	       then
% 		  try
% 		     {Show 'atleader '#@myident#@leader}
% 		     {Leaderref dummymsg()} %to force watcher to work!!
% 		     {Show 'atleader '#@myident#@leader}
		     
% 		  catch senderror
% 		  then
% 		     {Show @myident#'this client is dead'}
% 		  end
% 		  {self watcherproc(Leaderref)}
% % 		  Clintid={Getident Leaderref @identclient $}
% % 		  {Show @myident#'inside receivedcounter crashed client'#Clintid}
% % 		  MemUser={Memberuser Clintid @membership $}
% % 		  if {String.is MemUser $}==true
% % 		  then
% % 		     skip
% % 		  else
% % 		     skip
% % 		  end
% % 		  {Show'inside receivedcounter this is user'#@myident#MemUser#@membership#Clintid}
% % 		  aliveclients <- {List.subtract @aliveclients Leaderref $}
% % 		  identlist <- {List.subtract @identlist Clintid $}
% % 		  identclient <- {List.subtract @identclient Leaderref#Clintid $}
% % 		  membership <- {List.subtract @membership MemUser#Clintid $}
% % 		  {Show'failed clientid:inside receivedcounter '#Clintid#'myidentis'#@myident#'my identlist'#@identlist#@leader#@membership}
% % 		  {Show @myident#'listlengths'#{List.length @identlist  $}#{List.length @aliveclients $}}
% 		  %{self initiatselection()}
% 		  %{self watcherproc(Leaderref)}
	       %else
		%  {Show @myident#'inside receivedcounter this is not a member'}
	       %end
		  %{self watcherproc(Leaderref)}
	    % get Leader refrence!
	    % do watcher on that refrence!
	    end
	 end
      end
       %call from the new server to recovery
      meth recovery
	 {@user executeactionrecovery}
      end
     
      % call from old server to deinstall the watcher from him
      meth deinstalloldwatcher
	 try
	    {@server deinstallwatcher(@procwatcherserver Fault)}	    
	    {@server killserver()}
	    {Show @myident#'old server is killed'}
	 catch _  
	 then
	    skip
	 end
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Rule 33  Receive new state list of (objectid#state)
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth newstore(Newunlockedstore Newlockedstore Newserver Id)
	 % watcher procedure on newserver	
	 proc {Proc E FS}
	    if @connect==true
	    then
	       {self watcheronserver(permfail)}
	    else
	       {Show 'connection failed newstore'#@myident}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 proc {ProcT E FS}
	    {Show 'temporary failure detected newstore'}
	    if @connect==true
	    then
	       {self watcheronserver(tempfail)}
	    else
	       {Show 'connection failed newstore'}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 LL
      in
	 if @identnewstore==Id % valid message
	 then    
	    %status <- normal % make it the last thing!!!!
	    %listidstore <- nil
	    {Show @debug#@myident#'client:newstore'#Newunlockedstore#Newlockedstore}      
	    % Removes  all states 
	    {Dictionary.removeAll @objects}
	    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    % Set newstore  for locked objects
	    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    {ForAll Newlockedstore
	     proc {$ X}
		Objid State Obj in
		if X\=nil then		
		   X=Objid#State#_
		   try 
		      {Dictionary.get @classofobjects  Objid Obj#_#_}
		      {Dictionary.put @objects Objid State}
		      {Obj fromChunk(State)}
		   catch _ then % object creation when server failed
		      skip
		   end   		   
		else
		   skip
		end
	     end
	    }
	    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    % Set newstore  for unlocked objects
	    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    {ForAll Newunlockedstore
	     proc {$ X}
		Objid State Obj in
		if X\=nil then		
		   X=Objid#State     
		   
		   try 
		      {Dictionary.get @classofobjects  Objid Obj#_#_}
		      {Dictionary.put @objects Objid State}
		      {Obj fromChunk(State)}
		   catch _ then % object creation when server failed
		      skip
		   end
		else
		   skip
		end
	     end
	    }
	    {Show @myident#debug#'client:newstore:objects dict'#{Dictionary.entries  @objects $}}
	    
	    server <- Newserver
	 
	    % install a watcher on newserver
	    try
	       {@server installwatcher([permFail] Proc Fault)}
	       %{@server installtempwatcher(ProcT Fault)}
	       procwatcherserver <- Proc
	    catch cannotinstallwatcher
	    then
	       {Show @myident#installwatchererror}
	    end
	    if @leader==@myident % I am the leader
	    then
	       % ececute action recovery
	       {@user executeactionrecovery}
	    else
	       skip
	    end
	    iaminupdate <- true
            % forall objects in store
            {Dictionary.keys @objects LL}
           	
            {ForAll LL
             proc {$ Oi}
	         O in
	         {Dictionary.get @classofobjects  Oi O#_#_}
		%{@user executeactionupdate(O)}
		{@user executeactionupdate(Oi)} 
	     end
	    }
	   % {@user executeactionrecovery(P)}
	    iaminupdate <- false
	   
            % Client is unblocked	    
	    status <- normal
	    received <- 0
	    listidstore <- nil
	    newlockedstore <- nil
	    newunlockedstore <- nil
	    lockedobjects <- nil  
	    unlockedobjects <- nil
	    %ident <- 0
	    blocked <- unit
	    
	    % exute actionmemebership
	    {@user executeactionmembership(@membership)}
	    %{@user modifymembership(@membership)}
	    {Show 'actionmemvership after crating new server'#@myident#@membership}
%%% execute action
	    leader <- nil
	 else
	    skip
	 end
      end
      meth initiate_merge()
	 {@newserver initiate_merge(@server)}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Initiate a connection to the store
      % Client: new connected client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%      
      meth connectstore(Server Client User ?UserId ?Members ?Syn)	 
	 proc {Proc E FS}
	   % if connection not ended
	    if @connect==true
	    then
	       {self watcheronserver(permfail)}
	    else
	       % connection to afailed server
	       {Show 'connectstore failed server'#@myident}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 proc {ProcT E FS}
	    {Show 'temporary failure detected inside connection'}
	    if @connect==true
	    then
	       {self watcheronserver(tempfail)}
	    else
	       {Show 'connection failed temp'}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 Store
	 StoreIdelection
	 Idelection
	 Membership
      in
	 lock
	    client <- Client % Stationary Object
	    server <- Server % Stationary Object
	    user <- User %  Object
	    
       	   % set watcher on server
	    try
	       {Server installwatcher([permFail] Proc Fault)}
	       %{Server installtempwatcher(ProcT Fault)}
	       {Show @debug#@myident#'at client'#'watcherinstalledonserver'#Server}
	       procwatcherserver <- Proc
	    catch cannotinstallwatcher
	    then
	       {Show @myident#installwatchererror}
	       {Show @debug#@myident#'at client'#'cannotwatcherinstallonserver'#Server}
	    end
	    	 % connect to store
	    {Show @debug#@myident#'at client'#'connectstoreinit'#Store}
	    try {Server connectstore(@client @debug UserId Members StoreIdelection)}
	    catch senderror
	    then
	       %{@user raiseexception('cannotconnecttoserver')}
	       raise gs(connectionfailed) end 
	    end	   
	    {Show @debug#@myident#'at client'#'connectstoreinit'}
	    
	    StoreIdelection=Store#Idelection
	    idelection <- Idelection
	    membership <- Members
	    % verify that all variables in the store are binded
	    {ForAll Store
	     proc {$ X}
		Y Z T W in
		if X \= nil then
		   X=Y#Z#T#W
		   {Wait Y}{Wait Z}{Wait T}{Wait W}
		else skip
		end
	     end
	    }
	    %{self setident(UserId)}
	    {self setstore(Store)}
	    myident <- UserId
	    %leader <- @myident
	    leader <- nil
	    connect <- true
	    {Show @debug#@myident#'client'#'connectstorend'#Store}
	    Syn=unit
	    %{@user executeactionmembership(@membership)}
	    %{Show @debug#@myident#'insideconnect memebership'#@membership}
	    %{@user executeactionupdate(_)}
	    %{@user executeactionmembership(@membership)}
	    %{Show @debug#@myident#'insideconnect memebership'#@membership}
	 end
      end
      meth insideupdate(X)
	 X=@iaminupdate
      end
      meth setupdate
	iaminupdate <- true 
      end
      meth unsetupdate
	 iaminupdate <- false
      end
     % Call from the user
      meth disconnect(?Res)
	 Sync
	 proc{Getprocwatcher Cref Listproc ?Proc}
	    case Listproc of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Proc=H.2
	       else
		  {Getprocwatcher Cref Tail ?Proc}
	       end
	    end
	 end
	 %ProcWatch
      in
	 if @status==normal then
	    %disconnectstore(Client)
	    %{System.show 'Client:failed'}
	    try
	       {@server normaldisconnectstore(@client ?Sync)}
	       {Wait Sync}
	    catch senderror then raise gs(noservere)end
	    end
	    %{Show 'befor disable'}
	    %{Fault.defaultDisable _}
	    %{Show 'after disable'}
	    try
	      
	       {@server deinstallwatcher(@procwatcherserver Fault)}	    
	       %{@server killserver()}
	       {Show @debug#'at client'#'watcherdeinstallonserver'#@myident}
	    catch E  
	    then
	       {Show 'cannot deinstallwatcheron server'#@myident#E}
	    end
	     {ForAll @aliveclients
	      proc {$ Client}
		 ProcWatch
	      in
		 ProcWatch={Getprocwatcher Client @watcherprocclient $}
		 try 
		    {Client deinstallwatcher(ProcWatch Fault)}
		    {Show @debug#'at client'#'watcherdeinstallonclient'#@myident}
		   %{Delay 2000}
		 catch E
		 then
		    {Show deinstallwatchererror#@myident#E}
		 end
	      end
	    }
	    connect <- false
	    Res=true
	 else
	    {Show 'disconnection during consensus not allwed'}
	    Res=false
	 end
      end
      meth deinstall(C)
	 try 
	    {C deinstallwatcher(@watcherprocclient Fault)}
	    {Show @debug#'at clientdeinstall'#@myident#'watcherdeinstallonclient'}
	 catch E
	 then
	    {Show deinstallwatchererror#@myident#E}
	 end
      end
      
      meth getuniqueident(X)
	 try {@server getuniqueident(X)}
	 catch senderror then
	    raise gs(global_store_temporary_failed) end
	 end
      end
      meth getlocalref(Globalref LocalrefClass) %error
	 X in 
	 objwaitingref <- X
	 {Dictionary.get @classofobjects Globalref LocalrefClass#_#_}
	 %{Dictionary.condGet @classofobjects Globalref X  LocalrefClass}	 
      end      
      meth getglobalref(Localref ?Globalref)
	 try
	    {ForAll @localobjects
	     proc {$ O#Oid}
		if O==Localref then Globalref=Oid raise endforall end
		else
		   skip
		end
	     end
	    }
	 catch endforall
	 then skip
	 end	 
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% call from server
      %%%%%%%%%%%%%%%%%%%%%%%%%%%  
      meth addsetclient(Sc Idc Clientids OtherMember)
	%  proc{Proc M}
% 	    {self getmembership(M)}
% 	 end
%       in

	 {Show 'insideaddsetclient:watcherinstalleokclient:'#@myident}
	 aliveclients <- {List.flatten {List.append Sc @aliveclients $} $}
	 identlist <- {List.flatten {List.append Idc @identlist $} $}
	 identclient <- {List.flatten {List.append Clientids @identclient $} $}
	 %membership <- {List.flatten {List.append OtherMember @membership $} $}
	 {Show @debug#@myident#'at clientsetadded'#'aliveclient'#@aliveclients#@membership}
	 {Show @debug#@myident#'at clientsetadded'#'aliveclient'#@identlist#@identclient}
	 %Install watcher on each client of Sc
	 %{@user executeactionmembership(@membership)}
	 %{@user modifymembership(@membership)}
	 %{Show @debug#@myident#' inside adsetclient'#@membership} 
	 {ForAll @aliveclients
	  proc {$ Client}
	     proc {Proc E FS}
		if @connect==true then
		   {self watcherproc(Client)}
		else
		   {Show 'already disconnect addsetclient'#@myident}
		   skip
		end
	     end
	  in
	     try 
		{Client installwatcher([permFail] Proc Fault)}
		{Show @debug#@myident#'at client'#'watcherinstallonclient'#Client}
		watcherprocclient <- Client#Proc|@watcherprocclient
		
                %%% install a temp watcher
		%{Client installwatcher([tempFail] Proc Fault)}
	     catch cannotinstallwatcher
	     then
		{Show @myident#installwatchererror}
	     end
	     
	  end
	 }
	 %{@user executeactionmembership(@membership)}
	 %{Show @debug#'exectue membership inside adsetclient'} 
	 %{@user modifymembership(@membership)}
	 {Show @debug#@myident#' inside adsetclient'#@membership} 
      end
      meth mergeaddsetclient(Sc Idc Clientids OtherMember)
	%  proc{Proc M}
% 	    {self getmembership(M)}
% 	 end
%       in

	 {Show 'insideaddsetclient:watcherinstalleokclient:'#@myident}
	 aliveclients <- {List.flatten {List.append Sc @aliveclients $} $}
	 identlist <- {List.flatten {List.append Idc @identlist $} $}
	 identclient <- {List.flatten {List.append Clientids @identclient $} $}
	 membership <- {List.flatten {List.append OtherMember @membership $} $}
	 {Show @debug#@myident#'at clientsetadded'#'aliveclient'#@aliveclients#@membership}
	 {Show @debug#@myident#'at clientsetadded'#'aliveclient'#@identlist#@identclient}
	 %Install watcher on each client of Sc
	 %{@user executeactionmembership(@membership)}
	 %{@user modifymembership(@membership)}
	 %{Show @debug#@myident#' inside adsetclient'#@membership} 
	 {ForAll @aliveclients
	  proc {$ Client}
	     proc {Proc E FS}
		if @connect==true then
		   {self watcherproc(Client)}
		else
		   {Show 'already disconnect addsetclient'#@myident}
		   skip
		end
	     end
	  in
	     try 
		{Client installwatcher([permFail] Proc Fault)}
		{Show @debug#@myident#'at client'#'watcherinstallonclient'#Client}
		watcherprocclient <- Client#Proc|@watcherprocclient
		
                %%% install a temp watcher
		%{Client installwatcher([tempFail] Proc Fault)}
	     catch cannotinstallwatcher
	     then
		{Show @myident#installwatchererror}
	     end
	     
	  end
	 }
	 {@user executeactionmembership(@membership)}
	 %{Show @debug#'exectue membership inside adsetclient'} 
	 %{@user modifymembership(@membership)}
	 {Show @debug#@myident#' inside adsetclient'#@membership} 
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% call from server
      %%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth clientadded(C Ident UID)
	 User UserId 
	  proc {Proc E FS}
	    if @connect==true
	       then
	       {self watcherproc(C)}
	    else
	       {Show 'already disconnect clientadded'#@myident}
	       skip
	    end
	 end
      in
	 UID=User#UserId
	 {Show @debug#@myident#'at client'#'clientadded'#UID#User#UserId}
	% UID=User#UserId
        % setwatcher on each added client
	 try
	    {C installwatcher([permFail] Proc Fault)}
	    {Show @debug#@myident#'at clientadded'#'watcherinstallonclient'#C}
	    watcherprocclient <- C#Proc|@watcherprocclient
	 catch cannotinstallwatcher
	 then
	    {Show @myident#installwatchererror}
	 end
        % add client to the set of aliveclient
	 aliveclients <- C|@aliveclients
	 identlist    <- UserId|@identlist
	 identclient <-  C#UserId|@identclient
	 membership <- UID|@membership
         %{Dictionary.put @identclient Ident C}
	 {Show @debug#@myident#'at clientadded'#'aliveclient'#@aliveclients#@identlist#@identclient}
	 {@user executeactionmembership(@membership)}
	 {Show 'inside clientadded'#@debug#@myident#@membership}
	 %{@user modifymembership(@membership)}
      end
      
      %%%%%%%%%%%
      % call from server
      %%%%%%%%
      meth clientdisconnected(C Ident User)
	 proc{Getprocwatcher Cref Listproc ?Proc}
	    case Listproc of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Proc=H.2
	       else
		  {Getprocwatcher Cref Tail ?Proc}
	       end
	    end
	 end
	 ProcWatch
      in
	 ProcWatch={Getprocwatcher C @watcherprocclient $}
	 try 
	    {C deinstallwatcher(ProcWatch Fault)}
	    {Show @debug#'at client'#@myident#'watcherdeinstallonclient'#Ident}
	 catch E
	 then
	    {Show deinstallwatchererror#@myident#E#Ident}
	 end
	 aliveclients <-{List.subtract @aliveclients C $}
	 identlist <- {List.subtract @identlist Ident $}
	 identclient <- {List.subtract @identclient C#Ident $}
	 membership <- {List.subtract @membership User#Ident $}
	 {Show 'normaldisconnected clientid: '#Ident#'myidentis'#@myident#'my identlist'#@identlist#@leader#@membership}
	 {@user executeactionmembership(@membership)}
	 %{@user modifymembership(@membership)}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% List the store
      %%%%%%%%%%%%%%%%%%%%%%%%%%% 
      meth liststore() 
        % List all objects in store with their states
	{Show @debug#@myident#'client'#{Dictionary.entries @objects $}}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Add an object to the store call from user
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth addobjtostore(Client Obj Class Init)
	 {Wait @blocked} % read the paper for more comments
         % add an object to the store and broadcast it to all clients
	 N  State in
	 {NewName ?N}	 
	 localobjects <- Obj#N|@localobjects
	 %{SiteProperty.put N  Obj}
	 {Obj toChunk(?State)}
	 {Dictionary.put @objects  N State}
	 {Dictionary.put @classofobjects N Obj#Class#Init}
	 {Dictionary.put @counters N 0}
	 try {@server addobjtostore(N State Class Init Client)}
	 catch senderror
	 then
	    raise gs(global_store_temporary_failed) end
	 end
	 {Dictionary.put @dictbroadcast N true}
	 {Show @debug#@myident#'client:addobjtostore'#{Dictionary.entries  @objects $}}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Add an object to the store: call from client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth  addobj(Objid State Class Init)	 
	 Obj Stat in
	 lock
	    
	    {Show  @debug#@myident#'client:addobj'#Objid#State#Class#Init}
	    % test if object already added
	    {Show @debug#@myident#{Dictionary.entries @objects $}}	   
	    {Dictionary.condGet @objects Objid  noobj Stat}
	    if Stat==noobj then
	       {Show  @debug#@myident#'client:addobj:confirm'#Objid#State#Class#Init}
	    % create locally the object
	       Obj={New Class Init}
	       {Obj fromChunk(State)} % be careful State contains also feature globalref 
	    % update dictionaries      
	       {Dictionary.put @counters Objid 0}
	       localobjects <- Obj#Objid|@localobjects    
	       {Dictionary.put @objects  Objid State}
	       {Dictionary.put @classofobjects Objid Obj#Class#Init}
	       {Show  @debug#@myident#'client:addobj:user'#@user}    
	       {@user addtostore(Obj Class Init)}  
	       @objwaitingref=Obj
	       objwaitingref <- _
	       {Show  @debug#@myident#'client:addobj:end'#Objid#State#Class#Init}
	    else
	       skip
	    end
	 end
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Get the store
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth getstore(?Store ?Lasttransid ?Synch)
	 Lobj Temp Listobj in
	 lock @connectstore then
	    {Show @debug#@myident#'client:getstore'#Listobj}
	    
	    {Dictionary.entries @objects Lobj}
	    {List.map Lobj
	     fun {$ X}
	        O in
		if X \= nil then
		   Oclass Init Id State  in
		   X=Id#_
		   {Dictionary.get @classofobjects Id _#Oclass#Init}
		   {Dictionary.get @objects Id ?State}
		   
		   O={New Oclass Init}
		   {O fromChunk(State)}
		   %{Show 'etat'#{O show($)}}
		   Id#State#Oclass#Init
		else
		   nil
		end
	     end
	     ?Temp
	    }
	    
	    Listobj=Temp#@idelection
	    Store=Listobj
            % used for connection protocol
	    {Show @debug#@myident#'client:getstore'#Listobj}
	    Lasttransid= try {Dictionary.get @nametoint @lasttransid $}
			 catch _
			 then
			    none
			 end
	    
	    Synch=unit
	 end
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Set  the store call from the client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth setstore(ListObj) 
	lock
	 
	    {Show  @debug#@myident#'client:setstore'#ListObj}   
	    {Wait ListObj}	    
	    {ForAll ListObj
	     proc {$ Y}		
		Objid State Oclass Init in
		{Show @debug#@myident#'client:setstore'#Y}	
		Y=Objid#State#Oclass#Init
		{Wait  Objid}
		{Wait State}
		{Wait Oclass}
		{Wait Init}
		{self addobj(Objid State Oclass Init)}
	     end
	    }
	  
	end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Remove client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth disconnectstore(Client)
        % delete client 
	 try {@server disconnectstore(Client)}
	 catch senderror then
	    skip
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%% Set an ident to client
      %%%%%%%%%%%%%%%%%%%%%%%%%%% 
      meth setident(Idclient)
	 {Show @debug#@myident#'client:myident'#Idclient }
	 myident <- Idclient
	 leader <- @myident
	 %mymembership <- @debug#Idclient
	 {@user executeactionmymember(Idclient)} %lasupdate
      end
      meth ask(N Res)
	 {@server ask(N  ?Res)}
      end
      meth askpresent(N ?R)
	 {Dictionary.condGet @dictbroadcast N  false  R}
      end
      meth killserver()	 
	 try
	    {@server killserver()}
	    
	 catch senderror then
	    skip
	 end
	 
      end
      meth kill(End)
	 {Application.exit 0}
	 End=unit
      end
      meth getserver(?S)
	 S=@server
      end
      
      meth termination(?X)
	 try {@server termination(?X)}
	 catch senderror
	 then
	    skip
	 end
      end
      meth deletefromstore(Objid Res Synch)
	 Lock LocalrefClass in
	 {Wait @blocked}
	 {Show @myident#'client: send to server request lock for delete'}
	 try {@server request_locks_for_delete(Objid @client ?Lock ?Synch)}
	    {Wait Synch}
	 catch senderror % if send fails R18
	 then
	    {Show @myident#'senderror in deleteobject'}
	 end
	 if Lock == yes
	 then
	    
	    % you have the lock do what you want Res equal yes
	    {Show  @debug#@myident#'client:deleteobj:confirm'}
	    %{self getlocalref(Objid ?LocalrefClass)} % bug with getlocalref!!!!!!!
	    {Dictionary.get @classofobjects Objid LocalrefClass#_#_}
	    {Show 'client:object deleted before substract'#{List.length @localobjects}}
	    localobjects <- {List.subtract @localobjects LocalrefClass#Objid $} 
	    {Dictionary.remove @objects Objid}
	    {Dictionary.remove @classofobjects Objid}
	    {Dictionary.remove @counters Objid}
	    {Dictionary.remove @dictbroadcast Objid}
	    
	    {Show @myident#'client:object deleted'#{List.length @localobjects}}
	    Res=yes
	    
	 else
	    {Show @myident#'object is locked i cannt delete it'}
	    %{self setupdate}
	    %{@user executeactionabort(Objid)}
	    %{self unsetupdate}	
	    % You did not have the lock raise cannot remove this object Res =No
	 end
	 Synch=unit
	 
      end
      % call from the server
      meth delete_object(Objs)
	 Obj in
	 {Wait @blocked}
	 {Show  @debug#@myident#'client:deleteobj:confirm'}
	% here we must ask if this object exsists or not and wait him
         %{self getlocalref(Objs ?LocalrefClass)} % Bug there is a problem with this meth solve it.
	 {Show @myident#'get localref pass'}
	 {Dictionary.get @classofobjects Objs Obj#_#_}
	 localobjects <- {List.subtract @localobjects Obj#Objs $} 
	 {Dictionary.remove @objects Objs}
	 {Dictionary.remove @classofobjects Objs}
	 {Dictionary.remove @counters Objs}
	 {Dictionary.remove @dictbroadcast Objs}
	 {Show @debug#@myident#'client:object deleted'}
	 {self setupdate}
	 {@user executeactiondelete(Objs)}
	 {self unsetupdate}	
	 % send to the user to delete this objects idtosavedobjects
	 %{Dictionary.get @idtosavedobjects Objid ?State}			  
	 %{Dictionary.get @classofobjects Objid Obj#_#_}
	 %{Obj fromChunk(State)}
      end

%%%%%%%%%%%%%%%%% leave portocol    %%%%%%%%%%%%%%%%%%%%%

      meth leave(?Res)
	 Sync Clintid MemUser NewRef Pid F
	 proc{Getprocwatcher Cref Listproc ?Proc}
	    case Listproc of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Proc=H.2
	       else
		  {Getprocwatcher Cref Tail ?Proc}
	       end
	    end
	 end
	 
	 proc{Getident Cref Li ?Id}
	    case Li of nil then skip
	    []H|Tail then
	       if Cref==H.1 %H.1==Cref
	       then
		  Id=H.2
	       else
		  {Getident Cref Tail ?Id}
	       end
	    end
	 end
	 proc{Memberuser Id Li ?MemUser}
	    case Li of nil then skip
	    []H|Tail then
	       if Id==H.2 %H.1==Cref
	       then
		  MemUser=H.1
		  %{Show 'usermem='#MemUser}
	       else
		  {Memberuser Id Tail ?MemUser}
	       end
	    end
	 end
	 proc {Proc E FS}
	    if @connect==true
	    then
	       {self watcheronserver(permfail)}
	    else
	       {Show 'connection failed newstore'#@myident}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 proc {ProcT E FS}
	    {Show 'temporary failure detected newstore'}
	    if @connect==true
	    then
	       {self watcheronserver(tempfail)}
	    else
	       {Show 'connection failed newstore'}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 NewServer
      in
	 {Show @debug#'at client'#'starting leave by contacting old server'#@myident}
	 %if memebership length=0 then on leave
	 if @aliveclients\= nil then
	    if @status==normal then
	       status <- leaving
	       blocked <-_
	       try
		  {@server normaldisconnectstore(@client ?Sync)}% {@server leave(@client ?Sync) %send normaldisconnect instead of leav
		  {Wait Sync}
	       catch senderror then raise gs(noservere)end
	       end
	       try
		  
		  {@server deinstallwatcher(@procwatcherserver Fault)}	    
		  {Show @debug#'at client'#'watcherdeinstallonoldserver'#@myident}
	       catch E  
	       then
		  {Show 'cannot deinstallwatcheron oldserver'#@myident#E}
	       end
	       oldserver <- @server
	       connect <- false
	    %create new server dedicated to this client
	       {ForAll @aliveclients
		proc {$ Client}
		   ProcWatch
		in
		   ProcWatch={Getprocwatcher Client @watcherprocclient $}
		   try 
		      {Client deinstallwatcher(ProcWatch Fault)}
		      {Show @debug#'at client'#'watcherdeinstallonclient'#@myident}
		      Clintid={Getident Client @identclient $}
		      MemUser={Memberuser Clintid @membership $}
		      if {String.is MemUser $}==true
		      then
			 skip
		      else
			 skip
		      end
		      {Show'this is user'#MemUser#@membership#Clintid}
		      aliveclients<-{List.subtract @aliveclients Client $}
		      identlist <-{List.subtract @identlist Clintid $}
		      identclient <- {List.subtract @identclient Client#Clintid $}
		      membership <- {List.subtract @membership MemUser#Clintid $}
		   catch E
		   then
		      {Show deinstallwatchererror#@myident#E}
		   end
		end
	       }
	       @newserver=_
	       {Show 'creation newserver'#@newserver}
	       {NewStatObj ServerClass init(NewServer) NewServer _ _}
	       %{NewStatObj ServerClass init(@newserver) @newserver _ _} %its idclient =0 correct it
%%% Store emulator pid 
	       try
		  Pid={OS.getPID}
		  F={New Open.file init(name:'/tmp/pidserver' flags:[write create])}      
		  {F write(vs:Pid)}
		  {F close}
	       catch
		  E
	       then
		  {Show 'ErrorCreateFile'#E}
	       end
	       {Show debug#'client:newstore:objects dict'#{Dictionary.entries  @objects $}}
	    % set client set for the newserver
	    %{@newserver setclientset(@aliveclients @identlist @identclient @membership @client @newserver)}
	       server <- NewServer
	       %server <- @newserver
	       
	    % install a watcher on newserver
	       try
		  {@server installwatcher([permFail] Proc Fault)}
		  {@server installtempwatcher(ProcT Fault)}
		  procwatcherserver <- Proc
	       catch cannotinstallwatcher
	       then
		  {Show installwatchererror}
	       end
	       
	       
	    % for all current store change the current object id to new object id and keep the old one
	    % for later use in remerge operation  that is it !!!!!!
	    % send to the newserver your locked object to construct a new lock table
	    % send to the application the old server rfernce for remerge
	    % mark leave true
	    % end synchronization
	       
% 	    {Wait @blocked} % read the paper for more comments
%          % add an object to the store and broadcast it to all clients
	       % {ForAll @localobjects
% 		proc {$ O#Oid}
% 		   State  Class Init N
% 		in
% 		   {NewName ?N}
% 		   {Dictionary.get @objects Oid State}
% 		   {Dictionary.remove @objects Oid}
% 		   {Dictionary.put @objects N State}
% 		   {Dictionary.get @classofobjects Oid O#Class#Init}
% 		   {Dictionary.remove @classofobjects Oid}
% 		   {Dictionary.put @classofobjects N O#Class#Init}
% 		%{Dictionary.get @counters Oid Counter}
% 		   {Dictionary.remove @counters Oid}
% 		   {Dictionary.put @counters N 0}
% 		   % we should remove the old refernce from the user also !!!
% 		   {@user addtostore(O Class Init)}  
% 		   {@user newobjectref(O Oid N)} % this should be like new object not an modified one!!!!
% 		   mapedref <- N#Oid|@mapedref
% 		   templist <- O#N|@templist
% 		% replace the new one by the old one
% 		end
% 	       }
% 	       localobjects<- nil % i do'not know if this is correct!!!!!
% 	       localobjects <- @templist
% 	       templist<- nil
% 	       {Dictionary.keys @objects NewRef}
% 	       try
% 		  {@server newinformation(NewRef @client @myident @debug)} % for sending the new information about the local store
% 	       % listobjects <- Objid#Client|@listobjects
% 	       catch _
% 	       then
% 		  {Show nonewinformation}
% 	       end
	       connect <- true
	       leader <- @myident
	       aliveclients<- nil
	       identlist <- nil 
	       identclient <- nil
	       membership <-  nil
            % Client is unblocmakeked	    
	       @blocked=unit
	    % exute actionmemebership
	       {@user executeactionmembership(@membership)}
	    %{@user modifymembership(@membership)}
	       {Show 'actionmemvership after crating new server because of my leaving'#@myident#@membership}
	       leave <- true
	       Res=yes
	       status <- normal
	    else
	       {Show 'leaving during consensus not allwed'}
	    end
	 else
	    {Show 'leaving is not allowed'}
	    Res=no
	 end
      end
      meth merge(ProcMerge X)
	 proc{Getidref Id MapedList ?Currentid}
	    {ForAll MapedList
	     proc{$ Old#Common}
		if Id==Old
		then
		   Currentid=Common
		else
		   Currentid=notfound
		end
	     end}
	 end
	 proc {Proc E FS}
	   % if connection not ended
	    if @connect==true
	    then
	       {self watcheronserver(permfail)}
	    else
	       % connection to afailed server
	       {Show 'connectstore failed server'#@myident}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 proc {ProcT E FS}
	    {Show 'temporary failure detected'}
	    if @connect==true
	    then
	       {self watcheronserver(tempfail)}
	    else
	       {Show 'connection failed temp'}
	       %{Application.exit 0}
	       skip
	    end
	 end
	 OldStore Lobj
      in
	 % apply the send Proc to the old server and wait tell get new store from the old server
	 % 1- initiate connection to old server  %should guarentee that all current memeber agrre
	 % 2- send to aold server the current store with Proc
	 % 3- wait until having the new store new object states + oldstore with newstate
	 % 4- send to user all this information
	 % the old server will send to its leader the store recieved from the merged client and the Proc
	 % the leader will receive the mergeed store and compare it with its store,
	 % extract the newobject that is not exists in both store (new for store1 and new for store2)
	 % apply Proc foall the common objects in both store ( we can know the common object via old ref and ref)
	 % now ther is new client and new store sen this to the server and the server will propagate it.
	 % very good algorithm
	 % fiunally this client must change all its old id to be the current id again!!!!! very good
	 % the leader should solve the douplicated id client durnig leaving(Partitions)
	 try    
	   {@oldserver ping()}
	 catch senderror
	 then
	    {Show ' oldserver is dead no merge !!!!'} % we should not continue in this case
	 end
	 connect <- false % to start a connection with old server you should consider yourself disconnected
	 % do not forgit to deinstall the watcher from the current server
	 try
	    
	    {@server deinstallwatcher(@procwatcherserver Fault)}	    
	    {Show @debug#'at client'#'watcherdeinstallonserver'#@myident}
	 catch E  
	 then
	    {Show 'cannot deinstallwatcheron server'#@myident#E}
	 end
	 try
	    {@oldserver installwatcher([permFail] Proc Fault)}
	    {@oldserver installtempwatcher(ProcT Fault)}
	    {Show @debug#@myident#'at client'#'watcherinstalledonserver'}
	    procwatcherserver <- Proc
	 catch cannotinstallwatcher
	 then
	    {Show installwatchererror}
	    {Show @debug#@myident#'at client'#'cannotwatcherinstallonserver'}
	 end
	    	 % connect to store
	 {Dictionary.entries @objects Lobj}
	 {List.map Lobj
	  fun {$ X}
	     if X \= nil then
		Oclass Init Id State  in
		X=Id#_
		{Dictionary.get @classofobjects Id _#Oclass#Init}
		{Dictionary.get @objects Id ?State}
		
		%O={New Oclass Init} no need for this!!
		%{O fromChunk(State)}
		% we must send old id also but how?
		   %{Show 'etat'#{O show($)}}
		Id#State#Oclass#Init
	     else
		nil
	     end
	  end
	  ?OldStore
	 }
	 {Show @debug#@myident#'at client'#'mergestoreinit'#OldStore}
	 try {@oldserver mergestore(@client @debug @myident OldStore ProcMerge)}
	 catch senderror
	 then
	       %{@user raiseexception('cannotconnecttoserver')}
	    raise gs(connectionfailed) end 
	 end	   
	 %mapedref <- nil
	 server <- @oldserver
	 % set back the oldref to be the current ref and the current ref to the the old ref
	 % we have to take care of the new created object they have no old id !!!!
	 % {ForAll @localobjects
% 	  proc {$ O#Oid}
% 	     State  Class Init N
% 	  in
% 		  % {NewName ?N}
% 	     {Show 'rollback the old ref to be the newone start'}
% 	     %{@user getglobaloldref(Oid ?N)}
% 	     %mapedref
% 	     {Getidref Oid @mapedref ?N}
% 	     {Show 'rollback the old ref '#N}
% 	     if N \= notfound
% 	     then
% 		{Show 'rollback the old '}
% 		{Dictionary.get @objects Oid State}
% 		{Dictionary.remove @objects Oid}
% 		{Dictionary.put @objects N State}
% 		{Dictionary.get @classofobjects Oid O#Class#Init}
% 		{Dictionary.remove @classofobjects Oid}
% 		{Dictionary.put @classofobjects N O#Class#Init}
% 		%{Dictionary.get @counters Oid Counter}
% 		{Dictionary.remove @counters Oid}
% 		{Dictionary.put @counters N 0}
% 		{Show 'rollback the old '}
	
% 		%{@user addtostore(O Class Init)}
% 		%{@user newobjectref(O Oid N)} % this should be like new object not an modified one!!!!
% 		mapedref <- N#Oid|@mapedref
% 		templist <- O#N|@templist
% 		% replace the new one by the old one
% 		{Show 'rollback the old ref to be the newone end!!!'}

% 	     else
% 		{Show 'notfound'}
% 	     end
% 	  end
% 	 }
% 	 localobjects<- nil % i do'not know if this is correct!!!!!
% 	 localobjects <- @templist
% 	 {ForAll @localobjects
% 	  proc {$ O#Oid}
% 	       Class Init
% 	  in
% 		  % {NewName ?N}
% 	     {Show 'rollback the old ref to be the newone start'}
% 	     {Dictionary.get @classofobjects Oid O#Class#Init}
% 	     {@user addtostore(O Class Init)}
% 	     {@user newobjectref(O nil Oid)} % this should be like new object not an modified one!!!!
% 	  end
% 	 }
% 	 templist<- nil
% 	 mapedref <- nil
	 X=unit
      end
      % call from Servers
      meth propagatethis(ObjInfo)
	 Obj Oldref N State Class Init
      in
	 ObjInfo=N#State#Class#Init
	 {Show 'newobject state propagated'}
	 {Dictionary.get @classofobjects N Oldref#_#_}
	 %{Dictionary.remove @objects  N}
	 {Show 'oldref'#Oldref}
	 %Obj={New Class Init}
	 {Oldref fromChunk(State)}
	 {Show 'newobject created inside propagation'}
	 {Dictionary.remove @objects N}
	 {Dictionary.put @objects  N State}
	 {Dictionary.remove @classofobjects N}
	 {Dictionary.put @classofobjects N Oldref#Class#Init}
	 %localobjects <- {List.subtract @localobjects Oldref#N $}
	 %localobjects<- Obj#N|@localobjects
	 %{@user addtostore(Obj Class Init)}
	 %{@user newobjectref(Obj nil N)}
         % its update without transaction
	 iaminupdate <- true
	 {@user executeactionupdate(N)} %this is right
	 {Show 'propagarion done!!!'}
	 iaminupdate <- false
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % this will be run at the leader site and the most important method in this protocol
      % After the server locks all object the protocol becomes more flexible
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth remregestore(OldStore ProcMerge)
	 
	 proc{Getidref Id MapedList ?Currentid}
	    {ForAll MapedList
	     proc{$ Old#Common}
		if Id==Old
		then
		   Currentid=Common
		else
		   Currentid=notfound
		end
	     end}
	 end
	 StoreIdes
	 
      in
	 
	 {Show @debug#@myident#'at client remergestore begin'}
	 
	 % lock all objects
	 % compare between the old store and the current store
	 % any object on the old store have an old id = to one of the current obj id
	 % do merge acoording to the Proc 
	 % Union of oldstore and currentstore to apply the Proc
	 % the reminder of both will be treated as new objects !!!!!!!
	 {Dictionary.keys @classofobjects StoreIdes}
	 {ForAll OldStore
	  proc{$ Currentid#OldState#ObjClass#Init}
	     CurrentObj OldObj MergeObj  Init1 Oclass StateMergeObj in
	     % test whether this Oldoid is inside the NewOldRefList if so get its curreent object id
	     %    create new object as old one get the current object apply the proc to obtain the merged object
	     % else the its a new object and will be added as a new object in the merged store !!
	     
	     %{Getidref Oldoid NewOldRefList ?Currentid}
	     %{Show 'get current id form old id'#Currentid}
	     % if Currentid id bound then
	     if  {List.member Currentid StoreIdes $}==true then 
		{Show 'current id is returned'}
		{Dictionary.get @classofobjects Currentid CurrentObj#Oclass#Init1}
		OldObj ={New Oclass Init1}
		{OldObj fromChunk(OldState)}
		{Show 'current id is returned'#ProcMerge}
		%MergeObj={New Oclass Init1}
		%{Tmp fromChunk(OldState)}
		%{Show 'creat remerged object with oldstate'}
		%{Usermerge CurrentObj OldObj ?StateMergeObj}
		if {Procedure.is ProcMerge $}
		then
		   {Show {Procedure.arity ProcMerge $}}
		   try
		      {Show 'before running the proc'} 
		      %{Procedure.apply ProcMerge [CurrentObj OldObj StateMergeObj]}
		      {ProcMerge CurrentObj OldObj StateMergeObj}
		      {Show 'after running thr proc'} 
		   catch E then {Show error#E}
		   end
		end
		{Show 'after making merge'}
		MergeObj={New Oclass Init1}
		{MergeObj fromChunk(StateMergeObj)}
		%{MergeObj fromChunk(StateMergeObj)}
		{Show 'aftertochunk'}
		{Dictionary.remove @objects  Currentid}
		{Dictionary.put @objects  Currentid StateMergeObj}
		{Dictionary.remove @classofobjects Currentid}
		{Dictionary.put @classofobjects Currentid MergeObj#Oclass#Init1}
		localobjects <- {List.subtract @localobjects CurrentObj#Currentid $}
		localobjects<- MergeObj#Currentid|@localobjects
		%{@user addtostore(MergeObj Oclass Init1)}
		%{@user newobjectref(MergeObj nil Currentid)}
                % its update without transaction
		%{Show 'before execute action update'}
		iaminupdate <- true
		{@user executeactionupdate(Currentid)} %this is right 
		% send to server the new state to notify all other client and release look of that object because
		%{Show 'after execute action update'}
		try
		{@server serverpropagatethis(Currentid#StateMergeObj#Oclass#Init1 @client)}
		catch E then {Show error#E}end
		{Show 'after propagate this'}
		iaminupdate <- false
	     else
		{Show 'newobject created'}
		% create new object and add it to the store !!!!
		OldObj ={New ObjClass Init}
		{OldObj fromChunk(OldState)}
	        % update dictionaries      
		%{Dictionary.put @counters Oldoid OldObj}
		localobjects<- OldObj#Currentid|@localobjects    
		{Dictionary.put @objects Currentid  OldState}
		{Dictionary.put @classofobjects Currentid OldObj#ObjClass#Init}
		%{Dictionary.remove @counters Oid}
		{Dictionary.put @counters Currentid 0}
		   % we should remove the old refernce from the user also !!!
		{@user addtostore(OldObj ObjClass Init)}
		try {@server addobjtostore(Currentid OldState ObjClass Init @client)}
		catch senderror
		then
		   raise gs(global_store_temporary_failed) end
		end
		{Dictionary.put @dictbroadcast Currentid true}
		%{@user newobjectref(Obj Oid N)}
		% call the user for notify about newobject!!!!!
		% send to server this new object to broadcast it !!!!
		
	     end
	     
	  end
	 }
	 try {@server normalwork}
	 catch senderror
	 then
	    raise gs(global_store_temporary_failed) end
	 end
      end
      
      
   end
end

      




