%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Interface to the GS (Global Store Module)      %%
% Author: Ilis Alouini, Mostafa Al-Metwally     %%
% @copyright UCL & Region Wallone                %%
% Read LICENSE                                   %%
% Last update: 15/01/2001                        %%
% Documentation: http://www.ucl.ac.be/~metwally/ %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Create a new local store (new user)
%  User={New UserModule.userClass init(UrlServer UserName User)}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  O={NewObj Class Init}     %Create a new GS object (O is a sited  local reference)
%  {User addobjtostore(O OClass Init)} %Adding an Object to the store
%  {User getlocalref(Globalref ?O#Class#Init) % get the local reference of an object
%  {User getglobalref(O ?GlobalRef)  % get the global reference of an object        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  User <- {New UserClass init('http://..../serverticket')} -> Exception: connectionfailed
%  {User begin_trans(Procid Output ?Transid)}
%  {User user_update(Transid O Meth)} -> Exception:  noobjectinstore
%  {User end_trans(Transid)}
%  {User retrans(Procid ?ReTransid)}
%  {User newthread(Procid)}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Second Simpler API:  SEE the documentation 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
functor
import
   System
   Pickle 
   Connection
   ObjectSupport
   Application
   Fault
   ClientModule(clientClass:ClientClass) 
   StationaryObjectModule(newStatObj:NewStatObj)
   SiteProperty at 'x-oz://boot/SiteProperty'
export
   UserClass
   NewLocalStore   
define
   %%%%%%%%%%%%
   %% For Debug
   %%%%%%%%%%%%
   Debug=0
   
   proc {Show Message}
      if Debug==1 
      then
	 {System.show Message}
      else
	 skip
      end
   end
   %%%%%%%%%%%%
   %% uncomment the following line to set the siteproperty module
   \define  SitePropertyFlag
   %%%%%%%%%%%%
   %% skip procedure
   proc {Procskip X} skip end
   %%%%%%%%%%%%
   %% UserClass
   %%%%%%%%%%%%       
   class UserClass
      prop locking
      feat
	 username        % Username
	 userid          % UserId
      attr
	 client          % Client stationary object (asynchronous communication)
	 synchclient     % Client stationary object (synchronous communication)
	 objectclient    % Client object
	 user            % User Object
	 procids         % Dictionary: (Transid, Procid) 
	 transidtothreads% Dictionary: (Transid, Threadids)
	 statetrans      % Dictionary: transaction state (Transid,commit/abort)
	 threadtotransid:nil % List of (Threadid#Transid)
	 
	 synchvars:nil   % to synchronize threads inside transactions
	 synchvarsre:nil % to synchronize threads inside transactions
	 newlock:{NewLock $} % to serialize transactions
	 newlockrecom:{NewLock $} % to serialize recomputed transactions
	 synchlock:{NewLock $} % to serialize Transid with recomputation 
	 
	 currentthreadid % current Transid thread id
	 currenttransid  % current Transid transaction
	 currentoutput   % current Transid transaction output
	 currentprocid   % current transid procedure
	 
	 lastrecomputed  % last recomputed transaction identifier
	 recompute:false % true if recomputation in progress
	 resumedthreads:nil % list of resumedthreads
	 
	 usermember
	 store:nil       % List of local objects  (Refobj#Class#Init) in the local store
	 userstore:nil   % List of objects masquing local references (1 argument procedure)
	 globalrefmap  %dictionary(globalref, procedure#obj#class#Init)
	 membership    %List of membership Username#clientid
	 useraccount:nil  % username#userid
	 actionmembership:Procskip%proc{$ X} skip end
	 actioncreation:Procskip
	 actionupdates:proc {$ X} skip end
	 %actionabort:proc {$ X} skip end
	 actiondelete:proc {$ X}skip end
	 actionmigration:proc {$ X} skip end
	 actionrecovery:proc {$} skip end
	 iamintrans:false
	 siteproperty    % to simulate SiteProperty Module on prevoius Mozart versions
	 connection:false     %for connection/disconnection operation
      %%%%%%%%%% For debug
	 i:0             % Transactions counter 
	 block:unit      % used to block the user
	 leave:false     % for leave protocol
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% connect to client and server
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
      meth init(Server Username User KillFlag)
	  Client Synchclient ObjectClient X Members UserId in
	 self.username=Username
	 user <- self
         % Create a Client
	 {NewStatObj ClientClass init(Server Username) Client Synchclient ObjectClient}
	 client <- Client
	 synchclient <- Synchclient
	 objectclient <- ObjectClient
	 usermember <- Username
	 siteproperty<-{NewDictionary $}
	 globalrefmap<-{NewDictionary $}% added recently
         \ifdef SitePropertyFlag 
	   {SiteProperty.put user  self}
	   {SiteProperty.put oclient  @objectclient}
         \else
	     {Dictionary.put @siteproperty user  self}
	     {Dictionary.put @siteproperty oclient  @objectclient}  
	  \endif
	 {Show 'ObjectClient'#ObjectClient#Synchclient}
	 try    
	    {Server ping()}
	 catch senderror
	 then
	    {Show 'server reference is dead'}
	    %% the server reference is dead
	    %% will become alive after recovery
	    %% the user application should  retry the connection
	    raise gs(connectionfailed)	       
	    end
	 end
	 statetrans <-{NewDictionary $}
	 procids <-{NewDictionary $}     
	 transidtothreads<-{NewDictionary $}
	 
	 % Connect to the global store
	 {@objectclient connectstore(Server @client @user ?UserId ?Members X)}
	 %self.userid = UserId
	 {Wait X}
	 self.userid = UserId
	 connection <- true
	 {@user executeactionmembership(Members)}
	 {Show UserId#'insideconnect from user memebership'#Members}
	 if KillFlag==true then  
	       {Server killserver()}	    
	 else
	    skip
	 end	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\
      %%% Wait for the end of all Threads 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
      meth waitendthreads(Vars)
	 {Show self.userid#'begin'#waitendthreads(Vars)}
	 {ForAll Vars
	  proc {$ X}
	     {Wait X}
	  end
	 }
	 {Show self.userid#'end'#waitendthreads(Vars)}	
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Init a transaction 
      %%%% return transaction id
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
      meth begin_trans(Procid Output ?TransRes)
	 % Test If Procid is a procedure
	 Transid in
	 if @connection==true then
	    if {Procedure.is Procid}==true then
	       if {IsFree Transid}==false then raise gs(invalid_transid) end
	       else
		  thread
		     {Show self.userid#'IAMHERE'#{Thread.this $}}
		     lock @newlock then
			X  Y  in    
			synchvars <- X|@synchvars
			{Show self.userid#'IAMIN_NEWLOCK'#@synchvars}
			{Show self.userid#begintrans1#Procid#Transid#{Thread.this $}}
			currentoutput <- Output
			currentthreadid <- {Thread.this $}
			currentprocid <- Procid
			currenttransid <- Transid   
			{@objectclient begin_trans(?Transid)}		  
			{Show self.userid#begin_trans2#Procid#Transid#{Thread.this $}}
			{Dictionary.put @procids Transid Procid}
			{Dictionary.put @transidtothreads Transid [{Thread.this $}]}
			if {Dictionary.condGet @statetrans Transid noentry $}==noentry then 
			   {Dictionary.put @statetrans Transid _#Output#Y#0}
			else
			   skip
			end
			threadtotransid<- {Thread.this $}#Transid|@threadtotransid
			{Show self.userid#begintrans3#Procid#Transid#{Thread.this $}}
                        \ifdef SitePropertyFlag 
 			{SiteProperty.put localthreads @threadtotransid}
                        \else
			skip
                        \endif
			try
			   iamintrans <- true	
			   {Procid Y}
			   iamintrans <- false	
			catch endthread % this is raised from client: user_update
			then
			   iamintrans <- false	
			   {Show self.userid#catchedexception}
			   %raise objectdeleted end
			end
			X=unit
			{self waitendthreads(@synchvars)}
			{@objectclient end_trans(Transid)}
			synchvars <- nil
			{Show  self.userid#'SYNCHVARS'#@synchvars}
			{Show self.userid#endtrans3#Procid#Transid#{Thread.this $}}
			{self check_transaction(Transid TransRes _)}
			{Wait TransRes}
			threadtotransid <- {List.subtract @threadtotransid {Thread.this $}#Transid $}
			%\ifdef SitePropertyFlag 
			%{SiteProperty.put localthreads @threadtotransid}
			{Show self.userid#afterdealtinglocalthreads#@threadtotransid}
                        %\else
			%skip
                        %\endif
			
		     end
		  end
		  
	       end
	    else
	       raise gs(ill_formed_transaction) end
	    end
	 else raise gs(noconnection) end
	 end
      end
     %%%%%%%%% query%%%%%%%%
      meth query(Procid Output ?Transid)
	{Show 'this is query'} 
      end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Init a recomputed transaction 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
      meth transrecom(Procid Transid)
	 X  Y Out Res Z TransRes
	 proc {Getstatetrans}
	    try
	      {Dictionary.get @statetrans Transid Res#Out#_#Z}
	    catch _ then
	       {Show self.userid#'currentoutput'} % Transid killed before putting in dict statetrans 
	       {Dictionary.put @statetrans Transid _#@currentoutput#_#0}
	       Out=@currentoutput
	       Z=0
	   end
	 end
	 in
	 thread
	    lock @newlockrecom then
	       {Show self.userid#'transrecom'#Procid#Transid}
	       {Show self.userid#@objectclient}         
	       {@objectclient begin_transrecom(Transid)}	       
	       synchvarsre <- X |@synchvarsre		       
	       threadtotransid<-  {Thread.this $}#Transid|@threadtotransid
	       {Dictionary.put @transidtothreads Transid [{Thread.this $}]}
	        \ifdef SitePropertyFlag 
 		  {SiteProperty.put localthreads @threadtotransid}
 	       \else
 		  skip
	       \endif
	           
	       {Getstatetrans}
	       {Dictionary.remove @statetrans Transid}
	       {Dictionary.put @statetrans Transid Res#Out#Y#Z+1}
	       {Show self.userid#'transrecom'#Procid#Transid}          
	       try
		  iamintrans <- true
		  {Procid Y}
		  iamintrans <- false
	       catch
		  endthread
	       then
		  iamintrans <- false 
	       end
	       X=unit       
	       {self waitendthreads(@synchvarsre)} 
	       {@objectclient end_trans1(Transid)}
	       % return to normal calculation
	       synchvarsre <- nil       
	       if @lastrecomputed==Transid then		  	     
		   %  if {IsFree @suspendedthread $}==true
		    % then
			%{Show 'SUSPENDEDTHREAD_att'#@suspendedthread}
		    % else		  
			%{Show 'Resume_THREADS'#@suspendedthread}
			%resumedthreads <- @suspendedthread|@resumedthreads
                        % a corriger  suspendedthread devient une liste d ethread suspended
			%{Thread.resume @suspendedthread} 
		    % end
		     
		  recompute <- false
		  {@objectclient endrecom}
		  {Show self.userid#user#endrecompute}
	       else
		  skip
	       end
	       %{self check_transaction(Transid TransRes _)}
	       %{Wait TransRes}
	       threadtotransid <- {List.subtract @threadtotransid {Thread.this $}#Transid $}
	       {Show self.userid#'transrecom:end'#Procid#Transid#Res}
	    end
	 end	 	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% retrans operation 
      %%%% return transaction id
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
      meth retrans(Procid Output  ?Rtrid)
	 Y Res Out Z in
	 if @iamintrans==true then
	    {Show self.userid#'retrans:begin'}
	    if {List.member {Thread.this $} @resumedthreads $}==false
	    then
	       if {IsFree Rtrid $}==false then
		  {Dictionary.get @statetrans Rtrid Res#Out#_#Z}
		  {Dictionary.remove @statetrans Rtrid} % recomputation
		  {Dictionary.put @statetrans Rtrid Res#Out#Y#Z+1}
		  {@objectclient retrans(Rtrid)}
		  if {IsFree Rtrid $}==true then
		     {Application.exit 0}
		  else
		     skip
		  end
		  {self Retranscall(Procid Y  Rtrid)}
	       else
		  {@objectclient retrans(?Rtrid)}
		  {Dictionary.put @statetrans Rtrid _#Output#Y#0} % first call
		  {Dictionary.put @procids Rtrid Procid}     
		  {self Retranscall(Procid Y  Rtrid)}   
	       end
	    else
	       if {IsFree Rtrid $}==true then
		  {Show 'Currentthreadid'#{Thread.this $}}
		  Rtrid={NewName}
		  {Dictionary.put @statetrans Rtrid abort#Output#_#0} % first call
	      
	       else
		  skip
	       end
	       
	       {Show self.userid#'userignored retrans'}	 
	    end
	 else
	    raise gs(invalidcall) end
	 end
      end
      meth Retranscall(Procid Out Rtrid)
	 try {Procid Out}
	 catch endthread
	 then skip
	 end	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Update an object Obj  with method Meth
      %%%% within  transaction Transid 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
      meth user_update(Transid Obj Meth)
	 % precondition Obj is in the global store
	 {@objectclient user_update(Transid Obj  Meth)}	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Synchronous request of locks
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth waitlocks(Sel)	 	
	 Transid in
	 if @iamintrans==true
	 then
	    if @recompute==false then 
	       {self threadtotransid({Thread.this $} ?Transid)}
	       {@objectclient waitlocks(Transid Sel)}
	    else
	       {self threadtotransid({Thread.this $} ?Transid)}
	       {@objectclient waitlocks1(Transid Sel)}
	    end
	 else
	    raise gs(invalidcall) end
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Kill Transid threads
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% execute procid in a new thread
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth newthread(Procid)
	 X in
	 if @iamintrans==true then
	    if @recompute==false then 
	       synchvars <- X|@synchvars
	    else
	       synchvarsre <- X|@synchvarsre
	    end
	    thread
	       Listth  in
	       threadtotransid<-  {Thread.this $}#@currenttransid|@threadtotransid
	       {Dictionary.get @transidtothreads @currenttransid ?Listth}
	       {Dictionary.remove @transidtothreads @currenttransid}
	       {Dictionary.put @transidtothreads @currenttransid {List.append Listth {Thread.this $} $}}
	       try {Procid}
	       catch endthread
	       then skip
	       end
	       X=unit
	    end
	 else
	    raise gs(invalidcall) end
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% recompute transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth recompute(Trids)
	 %lock @synchlock then
	 
	 {Show self.userid#'user:recompute'#Trids}       
	 lastrecomputed <- {List.last Trids $}
	 % Kill the lastrecomputed threads'id : the other are forcely terminated
	 % {ForAll Trids	  
% 	     proc {$ Transid}
% 		{self killthreads(Transid)}
% 	     end
	     
% 	   }
	    
	 %end
	    recompute <- true	
	    {Show self.userid#'User:recompute'#begin}	    
	    {ForAll Trids	  
	     proc {$ Transid}
		Proc in
		try {Dictionary.get @procids Transid Proc}
		catch _ then
		   Proc=@currentprocid
	     end
		{Show self.userid#'User:recompute'#Transid#Proc}
		{self transrecom(Proc Transid)}	    
	     end
	    }
	 
      end     
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Notify to user committed transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth commit(Transid) 
	 Res  Out Y  in
	 i <- @i+1
	 {Dictionary.get @statetrans Transid Res#Out#Y#_}    
	 Res=commit
	 Out=Y
	 {Dictionary.remove @procids Transid}
	 %{Dictionary.remove @transidtothreads Transid}
	 {Dictionary.remove @statetrans Transid}
	 %{Dictionary.get @transidtothreads Transid Th}
	 {Dictionary.remove @transidtothreads Transid}
	 {Show self.userid#'Transaction '#Transid#'Committed'#@i#Out#{Dictionary.keys @statetrans}#{List.length @threadtotransid  $}}
	 %{@objectclient setupdate}
	 %{self executeactionupdate(_)}
	 %{@objectclient unsetupdate}
	 
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Notify to user aborted transaction
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth abort(Transid)
	 Res   in
	 i <- @i+1
	 {Show self.userid#'Transaction in abort'}
	 try {Dictionary.get @statetrans Transid Res#_#_#_}
	 catch E % not yet added to dictionary
	 then
	    {Show self.userid#' No Transactionid in abort'#E}
	    %{self abort(Transid)}
	    {Dictionary.put @statetrans Transid abort#_#_#_}
	 end
	 {Show self.userid#'Transaction '#Transid#'Aborted'#@i}
	 %{Dictionary.remove @statetrans Transid }
	 if {IsFree Res}
	 then Res=abort
	    {Show self.userid#'Transaction '#Transid#'AbortedEnd'#@i}
	    {Dictionary.remove @procids Transid}
	 %{Dictionary.remove @transidtothreads Transid}
	    {Dictionary.remove @statetrans Transid}
	 %{Dictionary.get @transidtothreads Transid Th}
	    {Dictionary.remove @transidtothreads Transid}
	 %{self killthreads(Transid)} %  (Bug with locks see ilalock1.oz ilalock2.oz)
	 else
	    {Show self.userid#itsafree}
	 end
      end
      meth affiche
        {Show self.userid#'comabo'#@i}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Check transaction 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth check_transaction(Transid Res Nbrec)	  
	 
	 {Wait Transid}	 
	 try
	    {Dictionary.get @statetrans Transid Res#_#_#Nbrec}
	    %{@objectclient info(Transid)}
	 catch _  % not yet put into dictionary -> retry
	 then
	    {self check_transaction(Transid Res Nbrec)}
	 end
      end            
      meth saveobj(Obj ?Name)
	 if @connection==true then
	    {Obj getglobalref(?Name)}
	    
	 %Name={@user getglobalref(Obj $)}
	    {Show name#Name}
	 %{Name={Obj getglobalref($)} }
	 else raise gs(noconnection)end
	 end
      end
	 
      meth loadobj(GlobalRef ?ProcLocalRef)
	 O  X  Result Thid Thid1 in
	    %{Connection.take {Pickle.load SF} ?GlobalRef}
	 if @connection ==true then
	    {@user getlocalref(GlobalRef X)}
	    {Show getlocalref#X}
	    % ask the server if the reference will come
	    % if never raise objectlost exception
	    %X=O#_#_
	    Thid={Thread.this}
	    thread
	    Thid1={Thread.this}
	       {@user  ask(GlobalRef ?Result)}
	       {Show ask#Result}
	       if Result==true then
		  skip
	       else % objectlost
		  {Thread.injectException Thid gs(objectlost)}  
	       end
	    end
	    %{Wait O} % wait until the object reference arrives
	    try {Thread.terminate Thid1} catch _ then skip end
	    {Dictionary.get @globalrefmap GlobalRef ProcLocalRef#_#_#_}
%{@user getobjproc(GlobalRef ?ProcLocalRef)}
	    {Show getobjproc#ProcLocalRef}
	 else raise gs(noconnection)end
	 end
      end
      
      meth unblock()
	 @block=unit
      end
      meth block()
	 block <- _
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Methods for connection
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Add objects to the store: call from user
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
      meth addobjtostore(Obj Class Init)
         % add an object to the store	 
	 if @connection==true then
	    store <- Obj#Class#Init|@store
	    {@objectclient addobjtostore(@client Obj Class Init)}
	    {Show self.userid#'user:addobjtostore'#Obj#Class#Init}
	 else raise gs(noconnection) end
	 end
      end
     %  meth modifymembership(Member)
% 	 lock
% 	    {Show 'insidemodifymembership'#Member}
% 	    membership <- Member
% 	    getmembership <- true
% 	    {Show 'insidemodifymembership'#Member}
% 	    {@user executeactionmembership(@membership)}
% 	    {Show self.username#@membership}
% 	 end
%       end
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%%% Add objects to the store: call from user
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth addobjtouserstore(N Pobj Obj Class Init)
	 lock
	    userstore <- {List.append @userstore [Pobj] $}
	    {Dictionary.put @globalrefmap N Pobj#Obj#Class#Init}
	    {Show self.userid#'user:addobjtouserstore'#Pobj#Obj#Class#Init}
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
      %%%% Add object to store: call from client
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%     
      meth addtostore(Obj Class Init)
	 Objp N in
	 {Show self.userid#'user:addtostorebegin'#N#Obj#Objp#Class#Init}
	 lock
	    
	    N={@user getglobalref(Obj $)}	    
	    % set features and attributes
	    {Obj setglobalref(N)}
	    \ifdef SitePropertyFlag 
	    {SiteProperty.put N  Obj}
	    {Show 'sitepropertyon'}
	    \else
	    {Dictionary.put @siteproperty N  Obj}
	    {Show 'sitepropertyon'}
	    \endif
	    store<- Obj#Class#Init|@store
	    {Show self.userid#'user:addtostorebegin'#N#Obj#Objp#Class#Init}
	    %{self getobjproc(N ?Objp)}
	    {Getobjproc N ?Objp}
	    {self addobjtouserstore(N Objp Obj Class Init)}
            %userstore <- Objp|@userstore
	    %{Dictionary.put @globalrefmap N Objp#Obj#Class#Init} % added recently
	    {Show self.userid#'user:@actioncreation'#@actioncreation#Obj}
	    {@objectclient setupdate}
	    {@actioncreation Objp}
	    {@objectclient unsetupdate}
	    {Show self.userid#'user:addtostoreend'#N#Obj#Objp#Class#Init}
	 end	    
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% List local store (local references)
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     %  meth liststore(Liststore)
% 	 GetList in
% 	 lock
% 	    fun{GetList L}
% 	       case L of
% 		  nil then nil
% 	       []  (Objs#_#_)|Res
% 	       then
% 		  Objs|{GetList Res}
% 	       end
% 	    end
% 	    Liststore={GetList @store}   
% 	    {Show self.userid#'liststore'#@store}	   
% 	 end
%       end
      meth liststore(Liststore)
	 GetList in
	 lock
	    fun{GetList L}
	       case L of
		  nil then nil
	       []  (Objs#_#_#_)|Res
	       then
		  Objs|{GetList Res}
	       end
	    end
	    Liststore={GetList @globalrefmap}   
	    {Show self.userid#'liststore'#@store}	   
	 end
      end
      
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% List local store (masquing local  references)
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth listuserstore(Liststore)
	  GetList in
	 lock
	    fun{GetList L}
	       case L of
		  nil then nil
	       []  Obj|Res
	       then
		  Obj|{GetList Res}
	       end
	    end

	    Liststore={GetList @userstore}   
	    {Show self.userid#'listuserstore'#@userstore}	   
	 end
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %% Delete object from the store (transactionally)
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth deleteobj(PObj ?Result)
	 Objid Res Synch Obj Class Init in
	 if @connection==true then
	    {Show self.userid#'user: begin delete object'}
	    {PObj getuniqueobjectid(?Objid)}
	    {@objectclient deletefromstore(Objid ?Res Synch)}
	    {Wait Synch}
	    if Res==yes
	    then
	       {Show self.userid#'user: reselt is yes'#{List.length @userstore}}
	       userstore <- {List.subtract @userstore PObj $}
	       {Dictionary.get @globalrefmap Objid PObj#Obj#Class#Init}
	       {Dictionary.remove @globalrefmap Objid}
	       store <- {List.subtract @store Obj#Class#Init  $}
	       {Show self.userid#'user:object deleted'#{List.length @userstore}#{List.length @store}#PObj#Obj#Class#Init}
	       Result=commit
	    %iam outside update
	    %set action of delete
	    else
	       {Show self.userid#'user:i cant delete this object'}
	       Result=abort
	   % {self executeactionabort(PObj)}
	    %raise exception that this object can't delete right now
	    end
	 else raise gs(noconnection)end
	 end
      end
      
      meth raiseexception(Error)
	 {Show self.userid#Error}
      end
      meth getlocalref(Globalref ?LocalrefClass)	
	 {@objectclient getlocalref(Globalref ?LocalrefClass)}
      end
      meth getglobalref(Localref ?Globalref)
	 {@objectclient getglobalref(Localref ?Globalref)} 
      end      
      meth killserver()	 
	 try {@objectclient killserver()}
	 catch senderror then
	    skip
	 end
      end     
     
      %%%%%%%%%%%%%%%%%%%
      % Test termination  
      %%%%%%%%%%%%%%%%%%%
      meth termination(?X)
	 lock @newlock then
	    {@objectclient termination(?X)}
	    {Wait X}
	    %{System.show termination}
	    %{System.show 'i:'#@i} 
	 end
	 {self affiche}
      end
      meth deletethread(Trid)
	 threadtotransid <- {List.subtract @threadtotransid _#Trid}
	 {Show self.userid#threaddeleted#{List.length @threadtotransid  $}}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Get class of a local reference
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      meth getclass(Lref ?C)
	 {ForAll @store
	  proc {$ Ref#Class#_}
	     if  Ref==Lref then C=Class
	     else
		skip
	     end
	  end
	 }
      end
      meth addsiteproperty(N O)
	 {Dictionary.put @siteproperty N O}
      end
      meth getserver(?S)
	 {@objectclient getserver(?S)}
      end
      meth getrecompute(X)
	 X=@recompute 
      end
      meth insideupdate(X)
	{@objectclient insideupdate(X)}
      end
      meth getstate(N ?State1)
	 {@objectclient getstate(N ?State1)}
      end
      meth insidelockok(X)
	{@objectclient insidelockok(X)}
      end
      meth disconnect(X)
	 X in
	 {@objectclient disconnect(X)}
	 {Wait X}
	 if X then
	 connection <- false
	 {Show self.userid#'user send to client disconnect'}
	 %{Application.exit 0}
	 %  \ifdef SitePropertyFlag 
% 	   {SiteProperty.remove user  self}
% 	   {SiteProperty.remove oclient  @objectclient}
%          \else
% 	     {Dictionary.remove @siteproperty user  self}
% 	     {Dictionary.remove @siteproperty oclient  @objectclient}  
% 	  \endif
	 else
	    skip
	 end
      end

      meth getuniqueident(X)
	 {@objectclient getuniqueident(X)}
      end
      meth ask( N Res)
	 {@objectclient ask(N Res)}
      end
      %%% For connection the store to external environment
      %%% To Do:  no transaction call inside action procedures
      
      meth executeactionupdate(Objid)
	 Proc in
	 % test si etat de l'objet modifie -> exception
	 {Show self.userid#debutexecuteactionupdates}
	 {Dictionary.get @globalrefmap Objid Proc#_#_#_}
	 {@actionupdates Proc}
	 {Show self.userid#finexecuteactionupdates#Proc}
      end
      meth executeactionabort(Objid)
	 Proc in
	 {@objectclient setupdate}
	 {Show self.userid#debutexecuteactionabort}
	 {Dictionary.get @globalrefmap Objid Proc#_#_#_}
	 {@actionabort Proc}
	 {Show self.userid#finexecuteactionabort}
	 {@objectclient unsetupdate}
      end
      meth executeactiondelete(Objid)
	 Procref Objref Class Init Keys1 Keys2 in
	 {Show self.userid#debutexecuteactiondelete}
	 {Dictionary.keys @globalrefmap Keys1}
	 {Show self.userid#'Keys'#Keys1}
	 {Dictionary.get @globalrefmap Objid Procref#Objref#Class#Init}
	 {Dictionary.remove @globalrefmap Objid}
	 {Dictionary.keys @globalrefmap Keys2}
	 {Show self.userid#'Keys'#Keys2#{List.length @userstore}#{List.length @store}}
	 userstore <- {List.subtract @userstore Procref $}
	 store <- {List.subtract @store Objref#Class#Init $}
	 {@actiondelete Procref}
	 {Show self.userid#finexecuteactiondelete#{List.length @userstore}#{List.length @store}#Procref#Objref#Class#Init}
      end
      meth executeactionrecovery
	 {Show self.userid#'actionrecovery begin'}
	 {@actionrecovery}
	 {Show self.userid#'actionrecovery end'}
      end
      meth executeactionmembership(Member)
	 {Show self.userid#'action membership begin'#Member}
	 {@actionmembership Member}
	 membership <- Member
	 %{self setactionmembership(P)}
	 {Show self.userid#'action membership end'#Member}
      end
      meth executeactionmymember(UsrId)
	 {Show UsrId#'mymembership begin'}
	 self.userid = UsrId
	 {Show UsrId#'mymembership end'}
      end
      meth setactioncreation(P)
	 if @actioncreation == Procskip
	 then
	     % first time
            %{@user executeactionmembership(_)}
	    {@objectclient setupdate}
	    {ForAll {self listuserstore($)}
	     proc {$ O}
		{P O}
	     end
	    }
	    {@objectclient unsetupdate}
	 else
	    skip
	 end
	 actioncreation <- P
      end
      meth setactionupdates(P)
	 actionupdates <- P
	 if {self listuserstore($)}==nil then skip
	 else
	    skip
	    %% bug how to recompute orders
	    %{@objectclient setupdate}
	    %{P _}
	    %{@objectclient unsetupdate}% recompute orders
	 end
      end
      meth setactionabort(P)
	 actionabort <- P
      end
      meth setactiondelete(P)
	 actiondelete <- P
	 if {self listuserstore($)}==nil then skip
	 else
	    skip
	    %% bug how to recompute orders
	    %{@objectclient setupdate}
	    %{P _}
	    %{@objectclient unsetupdate}% recompute orders
	 end
      end
      meth setactionrecovery(P)
	 actionrecovery <- P
      end
    
      meth setactionmembership(P)
	
	 if @actionmembership == Procskip
	 then
	    {P @membership}
	 else
	    skip
	 end
	 actionmembership <- P
      end
      meth setactionmigration(P)
	 actionmigration <- P
      end
      meth executeactionmigration()
	 {@actionmigration} 
      end
      meth getthread(?Threads ?Resumed  ?X)
	
	 \ifdef SitePropertyFlag 
	 try
	    {SiteProperty.get localthreads ?Threads}
	 catch _
	 then raise endforall end
	 end
                  \else
	 Threads=@threadtotransid 
	
\endif
	 Resumed=@resumedthreads
	 X=unit    
      end
%%%%%%%%%%%%%%%%% Leave and merge      %%%%%%%%%%%%
      meth leave(Res) %you should specify an action to be taken on both sides(run a procleave or not!!!
	 X in
	 % we should restrict it if you are the only user if memebership is nil !!!!!
	 {Show 'send leave to client'}
	 {@objectclient leave(X)}
	 {Wait X}
	 if X==yes
	    then
	    leave <- true
	    {Show 'client leaving successfully'}
	    Res=yes
	 else
	    {Show 'user:leave is not allowed'}
	    Res=no
	 end
      end
      meth merge(ProcMerge)
	 %send to client merge with old server with your store and Proc
	 X  in
	 {Show 'send merge to client'}
	 {@objectclient merge(ProcMerge X)}
	 {Wait X}
	 leave <- false
	 {Show 'client merge successfully'}
      end

      % Client Calls 
      meth newobjectref(O Oid N)
	 {O setglobalref(N)}
	 {O setoldref(Oid)}
	 
	 %{self getobjproc(N ?Objp)}
	 {Show 'end newobjet reference'}
      end
      %%%%%%%%%%%%%%%%%%%%%%%%% end leave and merge %%%%%%%%%%%%%%%%%%%%%%
   end
   %%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Wrap class of objects 
   %%%%%%%%%%%%%%%%%%%%%%%%%%%
   % proc {MakeClass Class  ?Generic}
%       class Generic from Class ObjectSupport.reflect
% 	 feat
% 	    globalref
% 	    delete % active, deleted
% 	 meth generic_init
% 	    skip
% 	 end
% 	 meth setglobalref(Gref)
% 	    self.globalref=Gref
% 	 end
% 	 meth setstate
% 	    self.delete=ok
% 	 end
%       end
%    end
   proc {MakeClass Class  ?Generic}
      class Generic from Class ObjectSupport.reflect
	 attr
	    globalref
	    oldref:nil
	 meth generic_init
	    skip
	 end
	 meth setglobalref(Gref)
	    %self.globalref=Gref
	    globalref <- Gref
	 end
	 meth setoldref(Old)
	    %self.oldref=Old
	    oldref <-Old
	 end
	 meth getglobalref(X)
	    X=@globalref
	 end
	 meth getoldref(X)
	    X=@oldref
	 end
      end
   end
   
       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Mask the local reference LocalObj  to an object 
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   proc {Getobjproc N ?ProcObj}
	 proc {Getthreadtotransid Thds Threadid Resumedthread ?Transid}	    
	    %{Show getthreadtotransid#Thds#Threadid#Transid}
	    {ForAll Thds
	     proc{$ Pair}
		Thid Trid in
		if Pair \= nil then
		   Pair=Thid#Trid
		   if Threadid==Thid
		   then
		      %if {List.member Trid @resumedthreads $}==false
		      if {List.member Trid Resumedthread $}==false
		      then
			 Transid=Trid raise stop end
		      else
			 {Show 'RAISE_FIN'} raise fin end
		      end
		   end
		end
	     end
	    }
	    raise endforall end   
	 end
      in
	 
	 proc{ProcObj Meth}
	    LocalObj Threads Trid LUser Resumedthread Oclient Syn
	 in
	    try
              \ifdef SitePropertyFlag 
	       {SiteProperty.get N ?LocalObj}
	       {SiteProperty.get user ?LUser}
	       {SiteProperty.get oclient ?Oclient}
	       % get state before apply the method from client
                \else         
	       %{Dictionary.get @globalrefmap N Procref#Objref#Class#Init} 
	       {Dictionary.get @siteproperty N ?LocalObj}
	       {Dictionary.get @siteproperty user ?LUser}
	       {Dictionary.get @siteproperty oclient ?Oclient}
	        % get state before apply the method from client
	       \endif
	    catch
	       E 
	    then
	       skip%{Show N#E}
	    end
	    case Meth of   
	       %getuniqueobjectid(?X) then  X=LocalObj.globalref
	       getuniqueobjectid(?X) then  X={LocalObj getglobalref($)}
	       else                      
	       try
		  {LUser getthread(?Threads ?Resumedthread Syn)}
		  {Wait Syn}
		  try {Getthreadtotransid Threads {Thread.this $} Resumedthread ?Trid}
		  catch
		     stop
		  then
		     if {LUser getrecompute($)}==true  then
			try
			   {Oclient user_update1(Trid N Meth)} % its brtten to send N
			catch endtrans then
			   raise endthread end
			end
		     else
			try
			   {Oclient user_update1(Trid N Meth)} 
			catch endtrans then
			   raise endthread end
			end
		     end
		  %[] fin then {Show 'UPDATE_OF_RESUMED_IGNORED'} raise endthread end
		  [] fin then skip raise endthread end 
		  end
	    catch 
		  endforall % transid not found
	       then
                   \ifdef SitePropertyFlag 		  
		  if {LUser insideupdate($)}==true %orelse {LUser insidelockok($)}==true
		  then
		     {LocalObj Meth}
		  else TempObj State1 in
		     TempObj=LocalObj % create new tempobject
		     {LUser getstate(N ?State1)}
		     {Wait State1}
		     %{LocalObj toChunk(State1)}
		     {TempObj fromChunk(?State1)}
		     {TempObj Meth}    % appply the meth on thae temp object
		     {TempObj fromChunk(?State1)}
		  end
\else		  
		  if {LUser insideupdate($)}==true 
		  then
		     {Show insidetrans}
		     {LocalObj Meth}
		     
		  else
		     TempObj State1 in
		     TempObj=LocalObj 
		     {LocalObj toChunk(?State1)}
		     {TempObj Meth}
		     {TempObj fromChunk(?State1)}
		  end
\endif
	       end
	    end
	 end 
   end
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % GlobalStore API Interface
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % Create a new local store
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   proc {NewLocalStore ServerRef Name ?UserId ?NewObjInStore ?LocalStore ?Movehere KillFlag}
      User StatUser      
   in    
      {NewStatObj UserClass  init(ServerRef Name User KillFlag) StatUser _  User }      
      %%%%% API interface to the localstore
      proc {LocalStore Meth}
	 case Meth of
	    newLocal(Module Y X Z T MH) then
	    F in
	    F= functor
	       import
		  GlobalStore(newLocalStore:NewLocalStore) at 'x-ozlib://metwally/dp/GlobalStore.ozf' 
	       export
	       define
		  try {NewLocalStore {StatUser getserver($)}  Y ?X ?Z ?T ?MH false}
		  catch senderror
		  then raise gs(connectionfailed) end
		  end
	       end
	    [_]={Module.apply [F]}
         %%% Cannot apply these methods on others LocalStore
	 %%% Hack: Introduce a ressource in procedure Sited: To find a better solution   
	 [] trans(Procid Out ?TransRes) then {User begin_trans(Procid Out ?TransRes)}
	 [] retrans(Procid ?Out ?Rtrid) then {User retrans(Procid ?Out ?Rtrid)}
	 [] newthread(Procid) then   {User  newthread(Procid)}
	 [] waitlocks(Sel) then {User waitlocks(Sel)}
	 [] termination(X) then  {User termination(X) }
	 [] disconnect(Res) then  {User disconnect(Res)}
	 %%% list objects 
	 [] liststore(L) then  {User liststore(L)}
	 %%% list external interface of objects
	 [] listglobalstore(L) then  {User listuserstore(L)}
	 [] deleteobj(Obj ?Res) then {User deleteobj(Obj ?Res)} % to be implemented
	 %%% interface to get the local object from an imported object
	 [] setnotifycreation(P) then  {User setactioncreation(P)}
	 [] setnotifyupdate(P) then  {User setactionupdates(P)}
	 %[] setnotifyabort(P) then  {User setactionabort(P)}
	 [] setnotifydelete(P) then {User setactiondelete(P)} % to be implemented
	 [] setnotifyrecovery(P) then  {User setactionrecovery(P)}
	 [] leave(Res) then {User leave(Res)}
	 [] merge(P) then {User merge(P)}
	 [] setnotifymembership(P) then {User setactionmembership(P)}
	 else
	    raise(gs(undefinedmethod)) end
	 end
      end
      UserId=User.userid
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Create a new object in the local store
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      proc {NewObjInStore Class Init ?ObjStore}
	 SClass
	 O  N  
      in
	 SClass={MakeClass Class}
	 O={New  SClass Init}    
	 % Add the object in the store; USER is a sited reference 
	 try 
	    {User addobjtostore(O SClass generic_init)}
	 catch _ then
	    {Show 'cannot create new object no connection'}
	    raise gs(noconnection) end
	 end      
	 % Add the local reference
	 N={User getglobalref(O $)}
         \ifdef SitePropertyFlag
	 {Show insidesiteproperty}
	 {SiteProperty.put N  O}
	 {Show objectaddedtositeproperrty}
            \else
	 {User addsiteproperty(N O)}
	    \endif
	 % Set features
	 {O setglobalref(N)}	 
         % Wrap the local reference 
	 {Getobjproc N ObjStore}
         % Add the  object global store reference to userstore 
	 {User addobjtouserstore(N ObjStore O SClass generic_init)}
      end
      
      
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      % Migrate the store
      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      proc {Movehere Module ?LS P}
	 F LS in
	 F= functor
	    import
	       GlobalStore(newLocalStore:NewLocalStore) at 'x-ozlib://metwally/dp/GlobalStore.ozf' 
	       export
	       define
	       try
		  {NewLocalStore {StatUser getserver($)} 'virtualmigrate'  _ ?LS _ true}
		  if {IsFree P}==false then {P}
		  else
		     skip
		  end
	       catch E
	       then raise E end
	       end
	       end
	 [_]={Module.apply [F]}
	 
      end
   end
   
end




